"""Tests for the Qblox Compiler class."""

import copy
import re
from unittest.mock import MagicMock

import numpy as np
import pytest
from qpysequence import Sequence
from qpysequence.constants import AWG_MAX_GAIN

from qililab.instruments.qblox import QbloxQCM, QbloxQRM
from qililab.pulse import Gaussian, Pulse, PulseBusSchedule, PulseSchedule, QbloxCompiler, Rectangular
from qililab.pulse.pulse_event import PulseEvent
from qililab.settings.digital.digital_compilation_bus_settings import DigitalCompilationBusSettings
from qililab.typings import AcquireTriggerMode, IntegrationMode, Parameter
from qililab.typings.enums import Line


class DummyQCM(QbloxQCM):
    """Dummy QCM class for testing"""

    _MIN_WAIT_TIME = 4

    def __init__(self, settings: dict):
        super().__init__(settings)
        self.device = MagicMock()


class DummyQRM(QbloxQRM):
    """Dummy QRM class for testing"""

    _MIN_WAIT_TIME = 4

    def __init__(self, settings: dict):
        super().__init__(settings)
        self.device = MagicMock(autospec=True)

@pytest.fixture(name="qcm_0")
def fixture_qcm_0():
    settings = {
        "alias": "qcm_0",
        "out_offsets": [0, 0.5, 0.7, 0.8],
        "awg_sequencers": [
            {
                "identifier": 0,
                "outputs": [0, 1],
                Parameter.IF.value: 100_000_000,
                Parameter.GAIN_I.value: 1,
                Parameter.GAIN_Q.value: 1,
                Parameter.GAIN_IMBALANCE.value: 0,
                Parameter.PHASE_IMBALANCE.value: 0,
                Parameter.OFFSET_I.value: 0,
                Parameter.OFFSET_Q.value: 0,
                Parameter.HARDWARE_MODULATION.value: False,
            },
            {
                "identifier": 1,
                "outputs": [0, 1],
                Parameter.IF.value: 100_000_000,
                Parameter.GAIN_I.value: 1,
                Parameter.GAIN_Q.value: 1,
                Parameter.GAIN_IMBALANCE.value: 0,
                Parameter.PHASE_IMBALANCE.value: 0,
                Parameter.OFFSET_I.value: 0,
                Parameter.OFFSET_Q.value: 0,
                Parameter.HARDWARE_MODULATION.value: False,
            },
        ],
    }
    return DummyQCM(settings=settings)

@pytest.fixture(name="qrm_0")
def fixture_qrm_0():
    settings = {
        "alias": "qrm_0",
        "out_offsets": [0.123, 1.23],
        "awg_sequencers": [
            {
                "identifier": 0,
                "outputs": [0, 1],
                Parameter.IF.value: 100_000_000,
                Parameter.GAIN_I.value: 1,
                Parameter.GAIN_Q.value: 1,
                Parameter.GAIN_IMBALANCE.value: 0,
                Parameter.PHASE_IMBALANCE.value: 0,
                Parameter.OFFSET_I.value: 0,
                Parameter.OFFSET_Q.value: 0,
                Parameter.HARDWARE_MODULATION.value: False,
                Parameter.SCOPE_ACQUIRE_TRIGGER_MODE.value: AcquireTriggerMode.SEQUENCER.value,
                Parameter.SCOPE_HARDWARE_AVERAGING.value: True,
                Parameter.SAMPLING_RATE.value: 1.0e09,
                Parameter.INTEGRATION_LENGTH.value: 2_123,
                Parameter.INTEGRATION_MODE.value: IntegrationMode.SSB.value,
                Parameter.SEQUENCE_TIMEOUT.value: 1,
                Parameter.ACQUISITION_TIMEOUT.value: 1,
                Parameter.HARDWARE_DEMODULATION.value: True,
                Parameter.SCOPE_STORE_ENABLED.value: True,
                Parameter.TIME_OF_FLIGHT.value: 40,
                Parameter.THRESHOLD.value: 0.5,
                Parameter.THRESHOLD_ROTATION.value: 45.0,
            },
            {
                "identifier": 1,
                "outputs": [0, 1],
                Parameter.IF.value: 200_000_000,
                Parameter.GAIN_I.value: 1,
                Parameter.GAIN_Q.value: 1,
                Parameter.GAIN_IMBALANCE.value: 0,
                Parameter.PHASE_IMBALANCE.value: 0,
                Parameter.OFFSET_I.value: 0,
                Parameter.OFFSET_Q.value: 0,
                Parameter.HARDWARE_MODULATION.value: False,
                Parameter.SCOPE_ACQUIRE_TRIGGER_MODE.value: AcquireTriggerMode.SEQUENCER.value,
                Parameter.SCOPE_HARDWARE_AVERAGING.value: True,
                Parameter.SAMPLING_RATE.value: 1.0e09,
                Parameter.INTEGRATION_LENGTH.value: 2_000,
                Parameter.INTEGRATION_MODE.value: IntegrationMode.SSB.value,
                Parameter.SEQUENCE_TIMEOUT.value: 1,
                Parameter.ACQUISITION_TIMEOUT.value: 1,
                Parameter.HARDWARE_DEMODULATION.value: True,
                Parameter.SCOPE_STORE_ENABLED.value: False,
                Parameter.TIME_OF_FLIGHT.value: 40,
                Parameter.THRESHOLD.value: 0.5,
                Parameter.THRESHOLD_ROTATION.value: 45.0,
            },
            {
                "identifier": 2,
                "outputs": [0, 1],
                Parameter.IF.value: 200_000_000,
                Parameter.GAIN_I.value: 1,
                Parameter.GAIN_Q.value: 1,
                Parameter.GAIN_IMBALANCE.value: 0,
                Parameter.PHASE_IMBALANCE.value: 0,
                Parameter.OFFSET_I.value: 0,
                Parameter.OFFSET_Q.value: 0,
                Parameter.HARDWARE_MODULATION.value: False,
                Parameter.SCOPE_ACQUIRE_TRIGGER_MODE.value: AcquireTriggerMode.SEQUENCER.value,
                Parameter.SCOPE_HARDWARE_AVERAGING.value: True,
                Parameter.SAMPLING_RATE.value: 1.0e09,
                Parameter.INTEGRATION_LENGTH.value: 2_000,
                Parameter.INTEGRATION_MODE.value: IntegrationMode.SSB.value,
                Parameter.SEQUENCE_TIMEOUT.value: 1,
                Parameter.ACQUISITION_TIMEOUT.value: 1,
                Parameter.HARDWARE_DEMODULATION.value: True,
                Parameter.SCOPE_STORE_ENABLED.value: False,
                Parameter.TIME_OF_FLIGHT.value: 40,
                Parameter.THRESHOLD.value: 0.5,
                Parameter.THRESHOLD_ROTATION.value: 45.0,
            },
        ],
    }
    return DummyQRM(settings=settings)

@pytest.fixture(name="qrm_1")
def fixture_qrm_1():
    settings = {
        "alias": "qrm_1",
        "out_offsets": [0.123, 1.23],
        "awg_sequencers": [
            {
                "identifier": 0,
                "outputs": [0, 1],
                Parameter.IF.value: 100_000_000,
                Parameter.GAIN_I.value: 1,
                Parameter.GAIN_Q.value: 1,
                Parameter.GAIN_IMBALANCE.value: 0,
                Parameter.PHASE_IMBALANCE.value: 0,
                Parameter.OFFSET_I.value: 0,
                Parameter.OFFSET_Q.value: 0,
                Parameter.HARDWARE_MODULATION.value: False,
                Parameter.SCOPE_ACQUIRE_TRIGGER_MODE.value: AcquireTriggerMode.SEQUENCER.value,
                Parameter.SCOPE_HARDWARE_AVERAGING.value: True,
                Parameter.SAMPLING_RATE.value: 1.0e09,
                Parameter.INTEGRATION_LENGTH.value: 2_123,
                Parameter.INTEGRATION_MODE.value: IntegrationMode.SSB.value,
                Parameter.SEQUENCE_TIMEOUT.value: 1,
                Parameter.ACQUISITION_TIMEOUT.value: 1,
                Parameter.HARDWARE_DEMODULATION.value: True,
                Parameter.SCOPE_STORE_ENABLED.value: True,
                Parameter.TIME_OF_FLIGHT.value: 40,
                Parameter.THRESHOLD.value: 0.5,
                Parameter.THRESHOLD_ROTATION.value: 45.0,
            }
        ],
    }
    return DummyQRM(settings=settings)

@pytest.fixture(name="buses")
def fixture_buses() -> dict[str, DigitalCompilationBusSettings]:
    return {
        "drive_q0": DigitalCompilationBusSettings(
            line=Line.DRIVE,
            qubits=[0]
        ),
        "flux_q0": DigitalCompilationBusSettings(
            line=Line.FLUX,
            qubits=[0]
        ),
        "readout_q0": DigitalCompilationBusSettings(
            line=Line.READOUT,
            qubits=[0],
        ),
        "readout_q1": DigitalCompilationBusSettings(
            line=Line.READOUT,
            qubits=[1],
        ),
        "readout_q2": DigitalCompilationBusSettings(
            line=Line.READOUT,
            qubits=[2],
        ),
        "readout_q3": DigitalCompilationBusSettings(
            line=Line.READOUT,
            qubits=[3],
        ),
    }

@pytest.fixture(name="bus_to_module_and_sequencer_mapping")
def fixture_bus_to_module_and_sequencer_mapping(qcm_0: DummyQCM, qrm_0: DummyQRM, qrm_1: DummyQRM):
    return {
        "drive_q0": {
            "module": qcm_0,
            "sequencer": qcm_0.get_sequencer(0)
        },
        "flux_q0": {
            "module": qcm_0,
            "sequencer": qcm_0.get_sequencer(1)
        },
        "readout_q0": {
            "module": qrm_0,
            "sequencer": qrm_0.get_sequencer(0)
        },
        "readout_q1": {
            "module": qrm_0,
            "sequencer": qrm_0.get_sequencer(1)
        },
        "readout_q2": {
            "module": qrm_0,
            "sequencer": qrm_0.get_sequencer(2)
        },
        "readout_q3": {
            "module": qrm_1,
            "sequencer": qrm_1.get_sequencer(0)
        }
    }


@pytest.fixture(name="qblox_compiler")
def fixture_qblox_compiler(buses, bus_to_module_and_sequencer_mapping):
    """Return an instance of Qblox Compiler class"""
    return QbloxCompiler(buses, bus_to_module_and_sequencer_mapping)


@pytest.fixture(name="pulse_bus_schedule")
def fixture_pulse_bus_schedule() -> PulseBusSchedule:
    """Return PulseBusSchedule instance."""
    pulse_shape = Gaussian(num_sigmas=4)
    pulse = Pulse(amplitude=0.8, phase=np.pi / 2 + 12.2, duration=50, frequency=1e9, pulse_shape=pulse_shape)
    pulse_event = PulseEvent(pulse=pulse, start_time=0, qubit=0)
    return PulseBusSchedule(timeline=[pulse_event], bus_alias="readout_q0")


@pytest.fixture(name="pulse_bus_schedule2")
def fixture_pulse_bus_schedule2() -> PulseBusSchedule:
    """Return PulseBusSchedule instance."""
    pulse_shape = Gaussian(num_sigmas=4)
    pulse = Pulse(amplitude=1, phase=0, duration=50, frequency=1e9, pulse_shape=pulse_shape)
    pulse_event = PulseEvent(pulse=pulse, start_time=0, qubit=1)
    return PulseBusSchedule(timeline=[pulse_event], bus_alias="readout_q1")


@pytest.fixture(name="pulse_bus_schedule_long_wait")
def fixture_pulse_bus_schedule_long_wait() -> PulseBusSchedule:
    """Return PulseBusSchedule instance."""
    pulse_shape = Gaussian(num_sigmas=4)
    pulse = Pulse(amplitude=0.8, phase=np.pi / 2 + 12.2, duration=50, frequency=1e9, pulse_shape=pulse_shape)
    pulse_event = PulseEvent(pulse=pulse, start_time=0, qubit=0)
    pulse_event2 = PulseEvent(pulse=pulse, start_time=200_000, qubit=0)
    return PulseBusSchedule(timeline=[pulse_event, pulse_event2], bus_alias="readout_q0")


@pytest.fixture(name="pulse_schedule_2qrm")
def fixture_pulse_schedule() -> PulseSchedule:
    """Return PulseBusSchedule instance."""
    pulse_event_0 = PulseEvent(
        pulse=Pulse(
            amplitude=0.8, phase=np.pi / 2 + 12.2, duration=50, frequency=1e9, pulse_shape=Gaussian(num_sigmas=4)
        ),
        start_time=0,
        qubit=2,
    )
    pulse_event_1 = PulseEvent(
        pulse=Pulse(amplitude=0.8, phase=0.1, duration=50, frequency=1e9, pulse_shape=Rectangular()),
        start_time=12,
        qubit=3,
    )
    return PulseSchedule(
        [
            PulseBusSchedule(timeline=[pulse_event_0], bus_alias="readout_q2"),
            PulseBusSchedule(timeline=[pulse_event_1], bus_alias="readout_q3"),
        ]
    )


@pytest.fixture(name="long_pulse_bus_schedule")
def fixture_long_pulse_bus_schedule() -> PulseBusSchedule:
    """Return PulseBusSchedule instance."""
    pulse = Pulse(amplitude=0.8, phase=np.pi / 2 + 12.2, duration=10**6, frequency=1e9, pulse_shape=Rectangular())
    pulse_event = PulseEvent(pulse=pulse, start_time=0, qubit=0)
    return PulseBusSchedule(timeline=[pulse_event], bus_alias="readout_q0")


# @pytest.fixture(name="pulse_schedule_odd_qubits")
# def fixture_pulse_schedule_odd_qubits() -> PulseSchedule:
#     """Returns a PulseBusSchedule with readout pulses for qubits 1, 3 and 5."""
#     pulse = Pulse(amplitude=1.0, phase=0, duration=1000, frequency=7.0e9, pulse_shape=Rectangular())
#     timeline = [PulseEvent(pulse=pulse, start_time=0, qubit=qubit) for qubit in [3, 1, 5]]
#     return PulseSchedule([PulseBusSchedule(timeline=timeline, port="feedline_input")])


def are_q1asm_equal(a: str, b: str):
    """Compare two Q1ASM strings and parse them to remove spaces, new lines and long_wait counters"""
    return "".join([cmd for cmd in a.strip().split() if "long_wait" not in cmd]) == "".join(
        [cmd for cmd in b.strip().split() if "long_wait" not in cmd]
    )


class TestQbloxCompiler:
    """Unit tests checking the QbloxQCM attributes and methods"""

    def test_amplitude_and_phase_in_program(self, qblox_compiler, pulse_bus_schedule):
        """Test that the amplitude and the phase of a compiled pulse is added into the Qblox program."""

        amplitude = pulse_bus_schedule.timeline[0].pulse.amplitude
        phase = pulse_bus_schedule.timeline[0].pulse.phase

        pulse_bus_schedule_qcm = copy.copy(pulse_bus_schedule)
        pulse_bus_schedule_qcm.bus_alias = "drive_q0"

        pulse_schedule = PulseSchedule([pulse_bus_schedule_qcm, pulse_bus_schedule])

        sequences = qblox_compiler.compile(pulse_schedule, num_avg=1, repetition_duration=2000, num_bins=1)
        program = list(sequences.items())[1][1][0]._program

        expected_gain = int(amplitude * AWG_MAX_GAIN)
        expected_phase = int((phase % (2 * np.pi)) * 1e9 / (2 * np.pi))

        bin_loop = program.blocks[2].components[1]

        assert bin_loop.components[0].args[0] == expected_gain
        assert bin_loop.components[0].args[1] == expected_gain
        assert bin_loop.components[1].args[0] == expected_phase

    def test_qrm_compile(self, qblox_compiler: QbloxCompiler, pulse_bus_schedule, pulse_bus_schedule2):
        """Test compile method."""
        pulse_schedule = PulseSchedule([pulse_bus_schedule])
        sequences = qblox_compiler.compile(pulse_schedule, num_avg=1000, repetition_duration=2000, num_bins=1)[
            "readout_q0"
        ]
        assert isinstance(sequences, list)
        assert len(sequences) == 1
        assert isinstance(sequences[0], Sequence)
        assert "Gaussian" in sequences[0]._waveforms._waveforms[0].name
        assert "Gaussian" in sequences[0]._waveforms._waveforms[1].name
        assert sum(sequences[0]._waveforms._waveforms[1].data) == 0
        assert len(sequences[0]._acquisitions._acquisitions) == 1
        assert sequences[0]._acquisitions._acquisitions[0].name == "acq_q0_0"
        assert sequences[0]._acquisitions._acquisitions[0].num_bins == 1
        assert sequences[0]._acquisitions._acquisitions[0].index == 0

        # test for different qubit, checkout that clearing the cache is working
        pulse_schedule2 = PulseSchedule([pulse_bus_schedule2])
        sequences = qblox_compiler.compile(pulse_schedule2, num_avg=1000, repetition_duration=2000, num_bins=1)[
            "readout_q1"
        ]
        assert isinstance(sequences, list)
        assert len(sequences) == 1
        assert isinstance(sequences[0], Sequence)
        assert len(qblox_compiler.module_and_sequencer_per_bus["readout_q1"]["module"].cache.keys()) == 2

    def test_qrm_compile_2qrm(
        self,
        qblox_compiler: QbloxCompiler,
        pulse_schedule_2qrm: PulseSchedule,
    ):
        """Test compile method for 2 qrms. First check a pulse schedule with 2 qrms, then one with
        only 1 qrm. Check that compiling the second sequence erases unused sequences in the unused qrm cache."""
        program = qblox_compiler.compile(pulse_schedule_2qrm, num_avg=1000, repetition_duration=2000, num_bins=1)

        assert len(program.items()) == 2
        assert "readout_q2" in program
        assert "readout_q3" in program
        assert len(qblox_compiler.module_and_sequencer_per_bus["readout_q2"]["module"].cache.keys()) == 1
        assert len(qblox_compiler.module_and_sequencer_per_bus["readout_q3"]["module"].cache.keys()) == 1

        assert list(qblox_compiler.module_and_sequencer_per_bus["readout_q2"]["module"].sequences.keys()) == [2]
        assert list(qblox_compiler.module_and_sequencer_per_bus["readout_q3"]["module"].sequences.keys()) == [0]

        assert len(program["readout_q2"]) == 1
        assert len(program["readout_q3"]) == 1

        sequences_0 = program["readout_q2"][0]
        sequences_1 = program["readout_q3"][0]

        assert isinstance(sequences_0, Sequence)
        assert isinstance(sequences_1, Sequence)

        assert "Gaussian" in sequences_0._waveforms._waveforms[0].name
        assert "Rectangular" in sequences_1._waveforms._waveforms[0].name

        q1asm_0 = """
            setup:
                            move             0, R0
                            move             1, R1
                            move             1000, R2
                            wait_sync        4

            start:
                            reset_ph
                            set_mrk          0
                            upd_param        4

            average:
                            move             0, R3
            bin:
                            set_awg_gain     26213, 26213
                            set_ph           191690305
                            play             0, 1, 4
                            wait             220
                            acquire          0, R3, 4
            long_wait_1:
                            wait             1772

                            add              R3, 1, R3
                            nop
                            jlt              R3, 1, @bin
                            loop             R2, @average
            stop:
                            set_mrk          0
                            upd_param        4
                            stop
        """

        q1asm_1 = """
            setup:
                            move             0, R0
                            move             1, R1
                            move             1000, R2
                            wait_sync        4

            start:
                            reset_ph
                            set_mrk          0
                            upd_param        4

            average:
                            move             0, R3
            bin:
            long_wait_2:
                            wait             12

                            set_awg_gain     26213, 26213
                            set_ph           15915494
                            play             0, 1, 4
                            wait             220
                            acquire          0, R3, 4
            long_wait_3:
                            wait             1760

                            add              R3, 1, R3
                            nop
                            jlt              R3, 1, @bin
                            loop             R2, @average
            stop:
                            set_mrk          0
                            upd_param        4
                            stop
        """
        seq_0_q1asm = repr(sequences_0._program)
        seq_1_q1asm = repr(sequences_1._program)

        assert are_q1asm_equal(q1asm_0, seq_0_q1asm)
        assert are_q1asm_equal(q1asm_1, seq_1_q1asm)

        # qblox modules 1 is the first qrm and 2 is the second
        assert qblox_compiler.module_and_sequencer_per_bus["readout_q2"]["module"].cache == {2: pulse_schedule_2qrm.elements[0]}
        assert qblox_compiler.module_and_sequencer_per_bus["readout_q3"]["module"].cache == {0: pulse_schedule_2qrm.elements[1]}
        assert qblox_compiler.module_and_sequencer_per_bus["readout_q2"]["module"].sequences == {2: sequences_0}
        assert qblox_compiler.module_and_sequencer_per_bus["readout_q3"]["module"].sequences == {0: sequences_1}

        # check that the qcm is empty since we didnt send anything to it
        assert not qblox_compiler.module_and_sequencer_per_bus["drive_q0"]["module"].cache
        assert not qblox_compiler.module_and_sequencer_per_bus["drive_q0"]["module"].sequences

    def test_long_wait_between_pulses(
        self, pulse_bus_schedule_long_wait: PulseBusSchedule, qblox_compiler: QbloxCompiler
    ):
        """test that a long wait is added properly between pulses if the wait time is longer than the max wait allowed by qblox"""
        expected_q1asm = """
        setup:
                        move             0, R0
                        move             1, R1
                        move             1000, R2
                        wait_sync        4

        start:
                        reset_ph
                        set_mrk          0
                        upd_param        4

        average:
                        move             0, R3
        bin:
                        set_awg_gain     26213, 26213
                        set_ph           191690305
                        play             0, 1, 4
                        wait             220
                        acquire          0, R3, 4
        long_wait_1:
                        move             3, R4
        long_wait_1_loop:
                        wait             65532
                        loop             R4, @long_wait_1_loop
                        wait             3396

                        set_awg_gain     26213, 26213
                        set_ph           191690305
                        play             0, 1, 4
                        wait             220
                        acquire          1, R3, 4
        long_wait_2:
                        move             3, R5
        long_wait_2_loop:
                        wait             65532
                        loop             R5, @long_wait_2_loop
                        wait             2956

                        add              R3, 1, R3
                        nop
                        jlt              R3, 1, @bin
                        loop             R2, @average
        stop:
                        set_mrk          0
                        upd_param        4
                        stop
        """

        pulse_schedule = PulseSchedule([pulse_bus_schedule_long_wait])
        program = qblox_compiler.compile(pulse_schedule, num_avg=1000, repetition_duration=400_000, num_bins=1)
        q1asm = repr(program["readout_q0"][0]._program)

        assert are_q1asm_equal(q1asm, expected_q1asm)

    # def test_qubit_to_sequencer_mapping(
    #     self, qblox_compiler: QbloxCompiler, pulse_schedule_odd_qubits: PulseSchedule, settings_6_sequencers: dict
    # ):
    #     """Test that the pulses to odd qubits are mapped to odd sequencers."""
    #     qblox_compiler.qblox_modules[1] = DummyQRM(settings=settings_6_sequencers)  # change qrm from fixture

    #     qblox_compiler.compile(
    #         pulse_schedule=pulse_schedule_odd_qubits, num_avg=1, repetition_duration=5000, num_bins=1
    #     )
    #     assert list(qblox_compiler.qblox_modules[1].cache.keys()) == [0, 2, 4]

    def test_acquisition_data_is_removed_when_calling_compile_twice(
        self, qblox_compiler: QbloxCompiler, pulse_bus_schedule
    ):  # FIXME: acquisition data should be removed at acquisition and not at compilation
        """Test that the acquisition data of the QRM device is deleted when calling compile twice."""
        pulse_event = PulseSchedule([pulse_bus_schedule])
        sequences = qblox_compiler.compile(pulse_event, num_avg=1000, repetition_duration=100, num_bins=1)[
            "readout_q0"
        ]
        qblox_compiler.module_and_sequencer_per_bus["readout_q0"]["module"].sequences = {
            0: sequences[0]
        }  # do this manually since we're not calling the upload method
        sequences2 = qblox_compiler.compile(pulse_event, num_avg=1000, repetition_duration=100, num_bins=1)[
            "readout_q0"
        ]
        assert len(sequences) == 1
        assert len(sequences2) == 1
        assert sequences[0] is sequences2[0]
        qblox_compiler.module_and_sequencer_per_bus["readout_q0"]["module"].device.delete_acquisition_data.assert_called_once_with(sequencer=0, all=True)

    def test_error_program_gt_repetition_duration(
        self, long_pulse_bus_schedule: PulseBusSchedule, qblox_compiler: QbloxCompiler
    ):
        """test that error is raised if circuit execution duration is longer than repetition duration"""
        pulse_schedule = PulseSchedule([long_pulse_bus_schedule])
        repetition_duration = 2000
        error_string = f"Circuit execution time cannnot be longer than repetition duration but found circuit time {long_pulse_bus_schedule.duration} > {repetition_duration} for qubit 0"
        with pytest.raises(ValueError, match=re.escape(error_string)):
            qblox_compiler.compile(pulse_schedule=pulse_schedule, num_avg=1000, repetition_duration=2000, num_bins=1)



"""Tests for the Pulse class."""

import itertools

import numpy as np
import pytest

from qililab.constants import PULSEEVENT
from qililab.pulse import (
    BiasTeeCorrection,
    Cosine,
    Drag,
    ExponentialCorrection,
    Gaussian,
    Pulse,
    PulseDistortion,
    PulseEvent,
    Rectangular,
)
from qililab.pulse.pulse_distortion.lfilter_correction import LFilterCorrection
from qililab.utils import Waveforms

# Parameters for the different Pulses
AMPLITUDE = [0.9]
PHASE = [0, np.pi / 3, 2 * np.pi]
DURATION = [48]
FREQUENCY = [0.7e9]
SHAPE = [
    Rectangular(),
    Cosine(),
    Cosine(lambda_2=0.3),
    Gaussian(num_sigmas=4),
    Drag(num_sigmas=4, drag_coefficient=1.0),
]

# Parameters for the different corrections.
TAU_BIAS_TEE = [1.3]
TAU_EXPONENTIAL = [0.9]
AMP = [-5.1, 0.8, 2.0]
NORM_FACTOR = 0.8


@pytest.fixture(
    name="pulse",
    params=[
        Pulse(amplitude=amplitude, phase=phase, duration=duration, frequency=frequency, pulse_shape=shape)
        for amplitude, phase, duration, frequency, shape in itertools.product(
            AMPLITUDE, PHASE, DURATION, FREQUENCY, SHAPE
        )
    ],
)
def fixture_pulses(request: pytest.FixtureRequest) -> Pulse:
    """Fixture for the Pulse class."""
    return request.param


@pytest.fixture(
    name="pulse_distortions",
    params=[
        [
            ExponentialCorrection(tau_exponential=tau_exponential, amp=amp),
            BiasTeeCorrection(tau_bias_tee=tau_bias_tee, norm_factor=NORM_FACTOR),
            LFilterCorrection(a=[0.7, 1.3], b=[0.5, 0.6]),
        ]
        for tau_exponential, amp, tau_bias_tee in itertools.product(TAU_EXPONENTIAL, AMP, TAU_BIAS_TEE)
    ],
)
def fixture_pulse_distortions(request: pytest.FixtureRequest) -> ExponentialCorrection:
    """Fixture for the pulse distortion class."""
    return request.param


@pytest.fixture(name="pulse_event")
def fixture_pulse_event() -> PulseEvent:
    """Load PulseEvent.

    Returns:
        PulseEvent: Instance of the PulseEvent class.
    """
    pulse_shape = Gaussian(num_sigmas=4)
    pulse = Pulse(amplitude=1, phase=0, duration=50, frequency=1e9, pulse_shape=pulse_shape)
    return PulseEvent(pulse=pulse, start_time=0)


class TestPulseEvent:
    """Unit tests checking the PulseEvent attributes and methods"""

    def test_modulated_waveforms_method(self, pulse_event: PulseEvent):
        """Test modulated_waveforms method."""
        waveforms = pulse_event.modulated_waveforms()

        assert waveforms is not None
        assert isinstance(waveforms, Waveforms)
        assert len(waveforms.values[0]) == len(waveforms.values[1]) == len(pulse_event.envelope())

    def test_envelope_method(self, pulse: Pulse, pulse_distortions: list[PulseDistortion]):
        """Test envelope method"""
        pulse_event = PulseEvent(pulse=pulse, start_time=0, pulse_distortions=pulse_distortions)

        resolution = 0.1

        envelope = pulse_event.envelope()
        envelope2 = pulse_event.envelope(resolution=resolution)
        envelope3 = pulse_event.envelope(amplitude=2.0, resolution=resolution)

        for env in [envelope, envelope2, envelope3]:
            assert env is not None
            assert isinstance(env, np.ndarray)

            if pulse_distortions:
                assert not np.array_equal(pulse.envelope(), env)

        # Test maximums
        if isinstance(pulse.pulse_shape, Cosine) and pulse.pulse_shape.lambda_2 > 0.0:
            # If lambda_2 > 0.0 the max amplitude is reduced
            assert round(np.max(np.abs(np.real(envelope))), 2 * int(np.sqrt(1 / 1.0))) < NORM_FACTOR * pulse.amplitude
            assert (
                round(np.max(np.abs(np.real(envelope2))), int(np.sqrt(1 / resolution))) < NORM_FACTOR * pulse.amplitude
            )
            assert round(np.max(np.abs(np.real(envelope3))), int(np.sqrt(1 / resolution))) < NORM_FACTOR * 2.0
            # If you check the form of this shape, the maximum never gets down 70% of the Amplitude for any lambda_2
            assert (
                round(np.max(np.abs(np.real(envelope))), 2 * int(np.sqrt(1 / 1.0)))
                > NORM_FACTOR * 0.7 * pulse.amplitude
            )
            assert (
                round(np.max(np.abs(np.real(envelope2))), int(np.sqrt(1 / resolution)))
                > NORM_FACTOR * 0.7 * pulse.amplitude
            )
            assert round(np.max(np.abs(np.real(envelope3))), int(np.sqrt(1 / resolution))) > NORM_FACTOR * 0.7 * 2.0
        else:
            # In the rest of pulse_shapes the maximum is the amplitude
            assert round(np.max(np.abs(np.real(envelope))), 2 * int(np.sqrt(1 / 1.0))) == round(
                NORM_FACTOR * pulse.amplitude, 5
            )
            assert round(np.max(np.abs(np.real(envelope2))), int(np.sqrt(1 / resolution))) == round(
                NORM_FACTOR * pulse.amplitude, 5
            )
            assert round(np.max(np.abs(np.real(envelope3))), int(np.sqrt(1 / resolution))) == round(
                NORM_FACTOR * 2.0, 5
            )

        assert len(pulse.envelope()) == len(envelope)
        assert len(envelope) * 10 == len(envelope2) == len(envelope3)

    def test_from_dict_method(self, pulse: Pulse, pulse_distortions: list[PulseDistortion]):
        """Test to_dict method"""
        pulse_event = PulseEvent(pulse=pulse, start_time=0, pulse_distortions=pulse_distortions)

        dictionary = pulse_event.to_dict()
        pulse_event2 = PulseEvent.from_dict(dictionary)

        dictionary2 = pulse_event2.to_dict()
        pulse_event3 = PulseEvent.from_dict(dictionary2)

        for event in [pulse_event2, pulse_event3]:
            assert event is not None
            assert isinstance(event, PulseEvent)

        assert pulse_event == pulse_event2 == pulse_event3

    def test_to_dict_method(self, pulse: Pulse, pulse_distortions: list[PulseDistortion]):
        """Test to_dict method"""
        pulse_event = PulseEvent(pulse=pulse, start_time=0, pulse_distortions=pulse_distortions)
        dictionary = pulse_event.to_dict()

        pulse_event2 = PulseEvent.from_dict(dictionary)
        dictionary2 = pulse_event2.to_dict()

        for dict_ in [dictionary, dictionary2]:
            assert dict_ is not None
            assert isinstance(dict_, dict)

        assert (
            dictionary
            == dictionary2
            == {
                PULSEEVENT.PULSE: pulse_event.pulse.to_dict(),
                PULSEEVENT.START_TIME: pulse_event.start_time,
                PULSEEVENT.PULSE_DISTORTIONS: [distortion.to_dict() for distortion in pulse_event.pulse_distortions],
                "qubit": None,
            }
        )

    def test_end_time(self, pulse: Pulse, pulse_distortions: list[PulseDistortion]):
        """Test end_time property."""
        pulse_event = PulseEvent(pulse=pulse, start_time=0, pulse_distortions=pulse_distortions)
        duration = pulse_event.duration

        assert duration is not None
        assert isinstance(duration, int)
        assert duration == pulse_event.end_time - pulse_event.start_time



import pytest

from qililab.settings.digital.digital_compilation_bus_settings import DigitalCompilationBusSettings
from qililab.typings.enums import Line


class TestDigitalCompilationBusSettings:
    def test_init_raises_error_if_weights_have_different_length(self):
        pass



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Driver for QDevil QDAC-II"""
from qcodes_contrib_drivers.drivers.QDevil.QDAC2 import QDac2 as Driver_QDac2

from qililab.typings.instruments.device import Device


class QDevilQDac2(Driver_QDac2, Device):
    """Typing class of the QDevil QDAC-II."""



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Pulse class."""

from dataclasses import dataclass

from qililab.constants import PULSE, RUNCARD
from qililab.pulse.pulse_shape.pulse_shape import PulseShape
from qililab.utils import Factory


@dataclass(frozen=True, eq=True)
class Pulse:
    """Describes a single pulse to be added to waveform array, which will be send through the lab buses to the chips.

    This class contains information about the pulse envelope shape, amplitude, phase, duration and frequency.

    Args:
        amplitude (float): Pulse's amplitude. Value normalized between 0 and 1.
        phase (float): Pulse's phase. Value in radians.
        duration (int): Pulse's duration. Value in ns.
        frequency (float): Pulse's frequency. Value in hertzs.
        pulse_shape (PulseShape): Pulse's envelope shape. Can be any of the different types of pulse shapes supported.
    """

    amplitude: float
    phase: float
    duration: int
    frequency: float
    pulse_shape: PulseShape

    def envelope(self, amplitude: float | None = None, resolution: float = 1.0):
        """Pulse 'envelope' property.

        Args:
            amplitude (float, optional): Pulse's amplitude. Value normalized between 0 and 1. Defaults to None
            resolution (float): Pulse's resolution. Value normalized between 0 and 1.

        Returns:
            list[float]: Amplitudes of the envelope of the pulse. Max amplitude is fixed to 1.
        """
        if amplitude is None:
            amplitude = self.amplitude
        return self.pulse_shape.envelope(duration=self.duration, amplitude=amplitude, resolution=resolution)

    @classmethod
    def from_dict(cls, dictionary: dict) -> "Pulse":
        """Loads Pulse object from dictionary.

        Args:
            dictionary (dict): Dictionary representation of the Pulse object.

        Returns:
            Pulse: Loaded class.
        """
        local_dictionary = dictionary.copy()
        pulse_shape_dict = local_dictionary[PULSE.PULSE_SHAPE]
        local_dictionary[PULSE.PULSE_SHAPE] = Factory.get(name=pulse_shape_dict[RUNCARD.NAME]).from_dict(
            pulse_shape_dict
        )

        return cls(**local_dictionary)

    def to_dict(self):
        """Returns dictionary of pulse.

        Returns:
            dict: Dictionary describing the pulse.
        """
        return {
            PULSE.AMPLITUDE: self.amplitude,
            PULSE.FREQUENCY: self.frequency,
            PULSE.PHASE: self.phase,
            PULSE.DURATION: self.duration,
            PULSE.PULSE_SHAPE: self.pulse_shape.to_dict(),
        }

    def label(self) -> str:
        """Returns short string representation of the Pulse object.

        Returns:
            str: String representation of the Pulse object.
        """
        return f"{self.pulse_shape!s} - {self.duration}ns"



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import ast
import os
from types import ModuleType

from IPython.core.magic import needs_local_scope, register_cell_magic
from IPython.core.magic_arguments import argument, magic_arguments, parse_argstring
from submitit import AutoExecutor

from qililab.config import logger

num_files_to_keep = 500  # needs to be a multiple of 4 and 5


def is_variable_used(code, variable):
    """Check whether any values are assigned to the output variable inside the magic cell."""
    tree = ast.parse(code)
    for node in ast.walk(tree):
        if isinstance(node, ast.Assign):
            for target in node.targets:
                if isinstance(target, ast.Name) and target.id == variable:
                    return True
    return False


@magic_arguments()
@argument(
    "-o",
    "--output",
    help="Output of the SLURM job. This name should correspond to a variable defined in the cell that we want to"
    " retrieve after execution. After queuing a cell, this variable will be converted to a `Job` class. To retrieve"
    " the results of the job, you need to call `variable.result()`.",
)
@argument("-p", "--partition", help="Name of the partition where you want to execute the SLURM job.")
@argument("-g", "--gres", default=None, help="GRES (chip) where you want to execute the SLURM job.")
@argument("-n", "--name", default="submitit", help="Name of the slurm job.")
@argument("-t", "--time", default=120, help="Time limit (in minutes) of the job.")
@argument(
    "-b",
    "--begin",
    default="now",
    help="Submit the batch script to the Slurm controller immediately, like normal, "
    "but tell the controller to defer the allocation of the job until the specified time. The time format can be"
    " either `HH:MM:SS`, `now+1hour`, `now+60minutes`, `now+60` (seconds by default), `2010-01-20T12:34:00`.",
)
@argument(
    "-l",
    "--logs",
    default=".slurm_job_data",
    help=(f"Path where you want slurm to write the logs for the last {num_files_to_keep} jobs."),
)
@argument(
    "-e",
    "--execution_environment",
    default=None,
    help="Select execution environment. Targets slurm by default but if '-e local' the job is run locally.",
)
@argument(
    "-lp",
    "--low_priority",
    default=None,
    help="By default lab jobs have higher priority than QaaS jobs. However, if '--lp True' or '--lp true' they will be queued with same priority as QaaS jobs, hence other Lab jobs will be executed first.",
)
@needs_local_scope
@register_cell_magic
def submit_job(line: str, cell: str, local_ns: dict) -> None:
    """Magic method that queues the content of a cell as a SLURM job.

    WARNING: Variables that start with an underscore (`_`) won't be recognized in the queued job.
    """
    # This method does NOT have a standard documentation because it corresponds to an Ipython magic method.

    args = parse_argstring(submit_job, line)
    output = args.output
    partition = args.partition
    gres = args.gres
    job_name = args.name
    time_limit = int(args.time)
    folder_path = args.logs
    execution_env = args.execution_environment
    begin_time = args.begin
    low_priority = args.low_priority

    nice_factor = 0
    if low_priority in ["True", "true"]:
        nice_factor = 1000000  # this ensures Lab jobs have 0 priority, same as QaaS jobs

    # Take all the import lines and add them right before the code of the cell (to make sure all the needed libraries
    # are imported inside the SLURM job)
    notebook_code = "\n".join(local_ns["In"]).split("\n")
    import_lines = [line for line in notebook_code if line.startswith(("import ", "from ")) and "slurm" not in line]
    executable_code = "\n".join([*import_lines, cell])

    # Create the executor that will be used to queue the SLURM job
    slurm_additional_parameters = {"begin": begin_time, "nice": nice_factor}
    if gres:
        slurm_additional_parameters |= {"gres": f"{gres}:1"}
    executor = AutoExecutor(folder=folder_path, cluster=execution_env)
    executor.update_parameters(
        slurm_partition=partition,
        name=job_name,
        timeout_min=time_limit,
        slurm_additional_parameters=slurm_additional_parameters,
    )

    # Compile the code defined above
    code = compile(executable_code, "<string>", "exec")
    # Take all the variables defined by the user in the Jupyter Notebook and the output variable
    variables = {
        k: v
        for k, v in local_ns.items()
        if not (
            isinstance(v, ModuleType) or k.startswith("_") or k in {"In", "Out", "exit", "quit", "open", "get_ipython"}
        )
    } | {output: None}

    # Define the function that will be queued as a SLURM job
    def function(code, variables):
        # Execute the code and return the output variable defined by the user
        exec(code, variables)  # noqa: S102
        return variables[output]

    # Check if output variables are defined or used in the magic cell
    if not is_variable_used(executable_code, output):
        raise ValueError(f"Output variable '{output}' was not assigned to any value inside the cell!")
    # Submit slurm job
    job = executor.submit(function, code, variables)
    logger.info("Your slurm job '%s' with ID %s has been queued!", job_name, job.job_id)
    # Overrides the output variable with the obtained job
    local_ns[output] = job

    # Delete info from past jobs
    job_ids = []
    file_paths = [os.path.join(folder_path, filename) for filename in os.listdir(folder_path)]
    for file_path in file_paths:
        try:
            job_ids.append(int(file_path.split("/")[1].split("_")[0]))

        # remove non-submitit files, not starting with an id
        except ValueError:
            logger.warning("%s shouldn't be in %s. It has been removed!", file_path.split("/")[1], folder_path)
            os.remove(file_path)

    for file_path in file_paths:
        if len(file_paths) >= num_files_to_keep and str(min(job_ids)) in file_path:
            os.remove(file_path)



"""Tests for the Qblox Module class."""
from typing import cast
from unittest.mock import MagicMock

import pytest
from qblox_instruments.qcodes_drivers.module import Module as QcmQrm
from qblox_instruments.qcodes_drivers.sequencer import Sequencer
from qpysequence import Acquisitions, Program, Sequence, Waveforms, Weights

from qililab.data_management import build_platform
from qililab.instruments.instrument import ParameterNotFound
from qililab.instruments.qblox import QbloxQRM
from qililab.platform import Platform
from qililab.qprogram.qblox_compiler import AcquisitionData
from qililab.typings import AcquireTriggerMode, IntegrationMode, Parameter


@pytest.fixture(name="platform")
def fixture_platform():
    """platform fixture"""
    return build_platform(runcard="tests/instruments/qblox/qblox_runcard.yaml")


@pytest.fixture(name="qrm")
def fixture_qrm(platform: Platform) -> QbloxQRM:
    qrm = cast(QbloxQRM, platform.get_element(alias="qrm"))

    sequencer_mock_spec = [
        *Sequencer._get_required_parent_attr_names(),
        "sync_en",
        "gain_awg_path0",
        "gain_awg_path1",
        "sequence",
        "mod_en_awg",
        "nco_freq",
        "scope_acq_sequencer_select",
        "channel_map_path0_out0_en",
        "channel_map_path1_out1_en",
        "demod_en_acq",
        "integration_length_acq",
        "mixer_corr_phase_offset_degree",
        "mixer_corr_gain_ratio",
        "connect_out0",
        "connect_out1",
        "connect_out2",
        "connect_out3",
        "marker_ovr_en",
        "offset_awg_path0",
        "offset_awg_path1",
        "connect_acq_I",
        "connect_acq_Q",
        "thresholded_acq_threshold",
        "thresholded_acq_rotation"
    ]

    module_mock_spec = [
        *QcmQrm._get_required_parent_qtm_attr_names(),
        "reference_source",
        "sequencer0",
        "sequencer1",
        "out0_offset",
        "out1_offset",
        "out2_offset",
        "out3_offset",
        "scope_acq_avg_mode_en_path0",
        "scope_acq_avg_mode_en_path1",
        "scope_acq_trigger_mode_path0",
        "scope_acq_trigger_mode_path1",
        "sequencers",
        "scope_acq_sequencer_select",
        "get_acquisitions",
        "disconnect_outputs",
        "disconnect_inputs",
        "arm_sequencer",
        "start_sequencer",
        "reset"
    ]

    # Create a mock device using create_autospec to follow the interface of the expected device
    qrm.device = MagicMock()
    qrm.device.mock_add_spec(module_mock_spec)

    qrm.device.sequencers = {
        0: MagicMock(),
        1: MagicMock(),
    }

    for sequencer in qrm.device.sequencers:
        qrm.device.sequencers[sequencer].mock_add_spec(sequencer_mock_spec)

    return qrm


class TestQbloxQRM:
    def test_init(self, qrm: QbloxQRM):
        assert qrm.is_awg()
        assert qrm.is_adc()
        assert qrm.alias == "qrm"
        assert len(qrm.awg_sequencers) == 2  # As per the YAML config
        assert qrm.out_offsets == [0.0, 0.1, 0.2, 0.3]
        sequencer = qrm.get_sequencer(0)
        assert sequencer.identifier == 0
        assert sequencer.outputs == [3, 2]
        assert sequencer.intermediate_frequency == 100e6
        assert sequencer.gain_imbalance == 0.05
        assert sequencer.phase_imbalance == 0.02
        assert sequencer.hardware_modulation is True
        assert sequencer.gain_i == 1.0
        assert sequencer.gain_q == 1.0
        assert sequencer.offset_i == 0.0
        assert sequencer.offset_q == 0.0

    def test_init_raises_error(self):
        with pytest.raises(ValueError):
            _ = build_platform(runcard="tests/instruments/qblox/qblox_qrm_too_many_sequencers_runcard.yaml")

    @pytest.mark.parametrize(
        "parameter, value",
        [
            # Test GAIN setting
            (Parameter.GAIN, 2.0),
            (Parameter.GAIN, 3.5),

            # Test GAIN_I and GAIN_Q settings
            (Parameter.GAIN_I, 1.5),
            (Parameter.GAIN_Q, 1.5),

            # Test OFFSET_I and OFFSET_Q settings
            (Parameter.OFFSET_I, 0.1),
            (Parameter.OFFSET_Q, 0.2),

            # Test IF setting (intermediate frequency)
            (Parameter.IF, 100e6),

            # Test HARDWARE_MODULATION setting
            (Parameter.HARDWARE_MODULATION, True),

            # Test GAIN_IMBALANCE setting
            (Parameter.GAIN_IMBALANCE, 0.05),

            # Test PHASE_IMBALANCE setting
            (Parameter.PHASE_IMBALANCE, 0.02),

            # Test OFFSET_OUT settings
            (Parameter.OFFSET_OUT0, 0.1),
            (Parameter.OFFSET_OUT1, 0.15),
            (Parameter.OFFSET_OUT2, 0.2),
            (Parameter.OFFSET_OUT3, 0.25),

            (Parameter.SCOPE_ACQUIRE_TRIGGER_MODE, "sequencer"),
            (Parameter.SCOPE_ACQUIRE_TRIGGER_MODE, "level"),
            (Parameter.SCOPE_HARDWARE_AVERAGING, True),
            (Parameter.SCOPE_HARDWARE_AVERAGING, False),
            (Parameter.SAMPLING_RATE, 0.09),
            (Parameter.HARDWARE_DEMODULATION, True),
            (Parameter.HARDWARE_DEMODULATION, False),
            (Parameter.INTEGRATION_LENGTH, 100),
            (Parameter.INTEGRATION_MODE, "ssb"),
            (Parameter.SEQUENCE_TIMEOUT, 2),
            (Parameter.ACQUISITION_TIMEOUT, 2),
            (Parameter.TIME_OF_FLIGHT, 80),
            (Parameter.SCOPE_STORE_ENABLED, True),
            (Parameter.THRESHOLD, 0.5),
            (Parameter.THRESHOLD_ROTATION, 0.5)
        ]
    )
    def test_set_parameter(self, qrm: QbloxQRM, parameter, value):
        """Test setting parameters for QCM sequencers using parameterized values."""
        qrm.set_parameter(parameter, value, channel_id=0)
        sequencer = qrm.get_sequencer(0)

        # Check values based on the parameter
        if parameter == Parameter.GAIN:
            assert sequencer.gain_i == value
            assert sequencer.gain_q == value
        elif parameter == Parameter.GAIN_I:
            assert sequencer.gain_i == value
        elif parameter == Parameter.GAIN_Q:
            assert sequencer.gain_q == value
        elif parameter == Parameter.OFFSET_I:
            assert sequencer.offset_i == value
        elif parameter == Parameter.OFFSET_Q:
            assert sequencer.offset_q == value
        elif parameter == Parameter.IF:
            assert sequencer.intermediate_frequency == value
        elif parameter == Parameter.HARDWARE_MODULATION:
            assert sequencer.hardware_modulation == value
        elif parameter == Parameter.GAIN_IMBALANCE:
            assert sequencer.gain_imbalance == value
        elif parameter == Parameter.PHASE_IMBALANCE:
            assert sequencer.phase_imbalance == value
        elif parameter == Parameter.SCOPE_ACQUIRE_TRIGGER_MODE:
            assert sequencer.scope_acquire_trigger_mode == AcquireTriggerMode(value)  # type: ignore[attr-defined]
        elif parameter == Parameter.INTEGRATION_LENGTH:
            assert sequencer.integration_length == value  # type: ignore[attr-defined]
        elif parameter == Parameter.SAMPLING_RATE:
            assert sequencer.sampling_rate == value  # type: ignore[attr-defined]
        elif parameter == Parameter.INTEGRATION_MODE:
            assert sequencer.integration_mode == IntegrationMode(value)  # type: ignore[attr-defined]
        elif parameter == Parameter.SEQUENCE_TIMEOUT:
            assert sequencer.sequence_timeout == value  # type: ignore[attr-defined]
        elif parameter == Parameter.ACQUISITION_TIMEOUT:
            assert sequencer.acquisition_timeout == value  # type: ignore[attr-defined]
        elif parameter == Parameter.TIME_OF_FLIGHT:
            assert sequencer.time_of_flight == value  # type: ignore[attr-defined]
        elif parameter in {Parameter.OFFSET_OUT0, Parameter.OFFSET_OUT1, Parameter.OFFSET_OUT2, Parameter.OFFSET_OUT3}:
            output = int(parameter.value[-1])
            assert qrm.out_offsets[output] == value

    def test_set_parameter_raises_error(self, qrm: QbloxQRM):
        """Test setting parameters for QCM sequencers using parameterized values."""
        with pytest.raises(ParameterNotFound):
            qrm.set_parameter(Parameter.BUS_FREQUENCY, value=42, channel_id=0)

        with pytest.raises(IndexError):
            qrm.set_parameter(Parameter.PHASE_IMBALANCE, value=0.5, channel_id=4)

        with pytest.raises(Exception):
            qrm.set_parameter(Parameter.PHASE_IMBALANCE, value=0.5, channel_id=None)

    @pytest.mark.parametrize(
        "parameter, expected_value",
        [
            # Test GAIN_I and GAIN_Q settings
            (Parameter.GAIN_I, 1.0),
            (Parameter.GAIN_Q, 1.0),

            # Test OFFSET_I and OFFSET_Q settings
            (Parameter.OFFSET_I, 0.0),
            (Parameter.OFFSET_Q, 0.0),

            # Test IF setting (intermediate frequency)
            (Parameter.IF, 100e6),

            # Test HARDWARE_MODULATION setting
            (Parameter.HARDWARE_MODULATION, True),

            # Test GAIN_IMBALANCE setting
            (Parameter.GAIN_IMBALANCE, 0.05),

            # Test PHASE_IMBALANCE setting
            (Parameter.PHASE_IMBALANCE, 0.02),

            # Test OFFSET_OUT settings
            (Parameter.OFFSET_OUT0, 0.0),
            (Parameter.OFFSET_OUT1, 0.1),
            (Parameter.OFFSET_OUT2, 0.2),
            (Parameter.OFFSET_OUT3, 0.3),

            (Parameter.SCOPE_ACQUIRE_TRIGGER_MODE, "sequencer"),
            (Parameter.SCOPE_HARDWARE_AVERAGING, True),
            (Parameter.SAMPLING_RATE, 1.e9),
            (Parameter.HARDWARE_DEMODULATION, True),
            (Parameter.INTEGRATION_LENGTH, 1000),
            (Parameter.INTEGRATION_MODE, "ssb"),
            (Parameter.SEQUENCE_TIMEOUT, 5.0),
            (Parameter.ACQUISITION_TIMEOUT, 1.0),
            (Parameter.TIME_OF_FLIGHT, 120),
            (Parameter.SCOPE_STORE_ENABLED, False),
            (Parameter.THRESHOLD, 1.0),
            (Parameter.THRESHOLD_ROTATION, 0.0)
        ]
    )
    def test_get_parameter(self, qrm: QbloxQRM, parameter, expected_value):
        """Test setting parameters for QCM sequencers using parameterized values."""
        value = qrm.get_parameter(parameter, channel_id=0)
        assert value == expected_value

    def test_get_parameter_raises_error(self, qrm: QbloxQRM):
        """Test setting parameters for QCM sequencers using parameterized values."""
        with pytest.raises(ParameterNotFound):
            qrm.get_parameter(Parameter.BUS_FREQUENCY, channel_id=0)

        with pytest.raises(IndexError):
            qrm.get_parameter(Parameter.PHASE_IMBALANCE, channel_id=4)

        with pytest.raises(Exception):
            qrm.get_parameter(Parameter.PHASE_IMBALANCE, channel_id=None)

    @pytest.mark.parametrize(
        "channel_id, expected_error",
        [
            (0, None),  # Valid channel ID
            (5, Exception),  # Invalid channel ID
        ]
    )
    def test_invalid_channel(self, qrm: QbloxQRM, channel_id, expected_error):
        """Test handling invalid channel IDs when setting parameters."""
        if expected_error:
            with pytest.raises(expected_error):
                qrm.set_parameter(Parameter.GAIN, 2.0, channel_id=channel_id)
        else:
            qrm.set_parameter(Parameter.GAIN, 2.0, channel_id=channel_id)
            sequencer = qrm.get_sequencer(channel_id)
            assert sequencer.gain_i == 2.0
            assert sequencer.gain_q == 2.0

    def test_initial_setup(self, qrm: QbloxQRM):
        """Test the initial setup of the QCM module."""
        qrm.initial_setup()

        # Verify the correct setup calls were made on the device
        qrm.device.disconnect_outputs.assert_called_once()
        for sequencer in qrm.awg_sequencers:
            qrm.device.sequencers[sequencer.identifier].sync_en.assert_called_with(False)

    def test_run(self, qrm: QbloxQRM):
        """Test running the QCM module."""
        qrm.sequences[0] = Sequence(program=Program(), waveforms=Waveforms(), acquisitions=Acquisitions(), weights=Weights())
        qrm.run(channel_id=0)

        sequencer = qrm.get_sequencer(0)
        qrm.device.arm_sequencer.assert_called_with(sequencer=sequencer.identifier)
        qrm.device.start_sequencer.assert_called_with(sequencer=sequencer.identifier)

    def test_upload_qpysequence(self, qrm: QbloxQRM):
        """Test uploading a QpySequence to the QCM module."""
        sequence = Sequence(program=Program(), waveforms=Waveforms(), acquisitions=Acquisitions(), weights=Weights())
        qrm.upload_qpysequence(qpysequence=sequence, channel_id=0)

        qrm.device.sequencers[0].sequence.assert_called_once_with(sequence.todict())

    def test_acquire_results(self, qrm: QbloxQRM):
        """Test uploading a QpySequence to the QCM module."""
        acquisitions_q0 = Acquisitions()
        acquisitions_q0.add(name="acquisition_q0_0")
        acquisitions_q0.add(name="acquisition_q0_1")

        acquisitions_q1 = Acquisitions()
        acquisitions_q1.add(name="acquisition_q1_0")
        acquisitions_q1.add(name="acquisition_q1_1")

        sequence_q0 = Sequence(program=Program(), waveforms=Waveforms(), acquisitions=acquisitions_q0, weights=Weights())
        sequence_q1 = Sequence(program=Program(), waveforms=Waveforms(), acquisitions=acquisitions_q1, weights=Weights())

        qrm.upload_qpysequence(qpysequence=sequence_q0, channel_id=0)
        qrm.upload_qpysequence(qpysequence=sequence_q1, channel_id=1)

        qrm.device.get_acquisitions.return_value = {
            "acquisition_q0_0": {
                "acquisition": {
                    "scope": {
                        "path0": {"data": [], "out-of-range": False, "avg_cnt": 0},
                        "path1": {"data": [], "out-of-range": False, "avg_cnt": 0},
                    },
                    "bins": {
                        "integration": {"path0": [1], "path1": [1]},
                        "threshold": [0],
                        "avg_cnt": [1],
                    },
                    "qubit": 0,
                    "measurement": 0,
                }
            },
            "acquisition_q0_1": {
                "acquisition": {
                    "scope": {
                        "path0": {"data": [], "out-of-range": False, "avg_cnt": 0},
                        "path1": {"data": [], "out-of-range": False, "avg_cnt": 0},
                    },
                    "bins": {
                        "integration": {"path0": [1], "path1": [1]},
                        "threshold": [0],
                        "avg_cnt": [1],
                    },
                    "qubit": 0,
                    "measurement": 0,
                }
            },
        }

        qrm.acquire_result()

        assert qrm.device.get_sequencer_status.call_count == 2
        assert qrm.device.get_acquisition_status.call_count == 2
        assert qrm.device.store_scope_acquisition.call_count == 1
        assert qrm.device.get_acquisitions.call_count == 2
        assert qrm.device.sequencers[0].sync_en.call_count == 1
        assert qrm.device.sequencers[1].sync_en.call_count == 1
        assert qrm.device.delete_acquisition_data.call_count == 2

    def test_acquire_qprogram_results(self, qrm: QbloxQRM):
        """Test uploading a QpySequence to the QCM module."""
        acquisitions = Acquisitions()
        acquisitions.add(name="acquisition_0")
        acquisitions.add(name="acquisition_1")

        sequence = Sequence(program=Program(), waveforms=Waveforms(), acquisitions=acquisitions, weights=Weights())
        qrm.upload_qpysequence(qpysequence=sequence, channel_id=0)

        qp_acqusitions = {
            "acquisition_0": AcquisitionData(bus="readout_q0", save_adc=False, shape=(-1,)),
            "acquisition_1": AcquisitionData(bus="readout_q0", save_adc=True, shape=(-1,))
        }

        qrm.acquire_qprogram_results(acquisitions=qp_acqusitions, channel_id=0)

        assert qrm.device.get_acquisition_status.call_count == 2
        assert qrm.device.store_scope_acquisition.call_count == 1
        assert qrm.device.get_acquisitions.call_count == 2
        assert qrm.device.delete_acquisition_data.call_count == 2

    def test_clear_cache(self, qrm: QbloxQRM):
        """Test clearing the cache of the QCM module."""
        qrm.cache = {0: MagicMock()}  # type: ignore[misc]
        qrm.clear_cache()

        assert qrm.cache == {}
        assert qrm.sequences == {}

    def test_reset(self, qrm: QbloxQRM):
        """Test resetting the QCM module."""
        qrm.reset()

        qrm.device.reset.assert_called_once()
        assert qrm.cache == {}
        assert qrm.sequences == {}


# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# mypy: disable-error-code="attr-defined"
import os
from dataclasses import dataclass
from datetime import datetime
from typing import Any

import h5py
import matplotlib.pyplot as plt
import numpy as np


@dataclass
class DimensionInfo:
    """Dataclass to store information of a Variable"""

    labels: list[str]
    values: list[np.ndarray]


class ExperimentResults:
    """Provides methods to access the experiment results stored in an HDF5 file."""

    QPROGRAMS_PATH = "qprograms"
    MEASUREMENTS_PATH = "measurements"
    RESULTS_PATH = "results"
    VARIABLES_PATH = "variables"
    EXPERIMENT_PATH = "experiment"
    PLATFORM_PATH = "platform"
    EXECUTED_AT_PATH = "executed_at"
    EXECUTION_TIME_PATH = "execution_time"

    S21_PLOT_NAME = "S21.png"

    def __init__(self, path: str):
        """Initializes the ExperimentResults instance.

        Args:
            path (str): The file path to the HDF5 results file.
        """
        self.path = path
        self.data: dict[tuple[str, str], Any] = {}  # To hold links to the data of the results for in-memory access
        self.dimensions: dict[tuple[str, str], Any] = (
            {}
        )  # To hold links to dimensions of the results for in-memory access
        self._file: h5py.File

    def __enter__(self):
        """Opens the HDF5 file for reading.

        Returns:
            ExperimentResults: The ExperimentResults instance.
        """
        self._file = h5py.File(self.path, mode="r")

        # Prepare access to each results dataset and its dimensions
        for qprogram_name in self._file[ExperimentResults.QPROGRAMS_PATH]:
            qprogram_data = self._file[f"{ExperimentResults.QPROGRAMS_PATH}/{qprogram_name}"]
            for measurement_name in qprogram_data[ExperimentResults.MEASUREMENTS_PATH]:
                self.data[qprogram_name, measurement_name] = qprogram_data[
                    f"{ExperimentResults.MEASUREMENTS_PATH}/{measurement_name}/results"
                ]
                # Store the dimensions
                self.dimensions[qprogram_name, measurement_name] = qprogram_data[
                    f"{ExperimentResults.MEASUREMENTS_PATH}/{measurement_name}/results"
                ].dims

        return self

    def __exit__(self, *_):
        """Exit the context manager and close the HDF5 file."""
        if self._file is not None:
            self._file.close()

    def get(self, qprogram: int | str = 0, measurement: int | str = 0) -> tuple[np.ndarray, list[DimensionInfo]]:
        """Retrieves data and dimensions for a specified quantum program and measurement.

        Args:
            qprogram (int | str, optional): The index or name of the quantum program. Defaults to 0.
            measurement (int | str, optional): The index or name of the measurement. Defaults to 0.

        Returns:
            tuple[np.ndarray, list[dict]]: A tuple containing the data array and a list of dimension dictionaries.
        """
        if isinstance(qprogram, int):
            qprogram = f"QProgram_{qprogram}"
        if isinstance(measurement, int):
            measurement = f"Measurement_{measurement}"

        data = self.data[qprogram, measurement][()]
        dims = [
            DimensionInfo(labels=dim.label.split(","), values=[values[()] for values in dim.values()])
            for dim in self.dimensions[qprogram, measurement]
        ]

        return data, dims

    def __getitem__(self, key: tuple):
        """Get an item from the results dataset.

        Args:
            key (tuple): A tuple of (qprogram_name, measurement_name, *indices).

        Returns:
            float: The value stored in the results dataset at the given indices.
        """
        qprogram_name, measurement_name, *indices = key
        if isinstance(qprogram_name, int):
            qprogram_name = f"QProgram_{qprogram_name}"
        if isinstance(measurement_name, int):
            measurement_name = f"Measurement_{measurement_name}"
        return self.data[qprogram_name, measurement_name][tuple(indices)]

    def __len__(self):
        """Gets the total number of results datasets.

        Returns:
            int: The number of results datasets.
        """
        return len(self.data)

    def __iter__(self):
        """Gets an iterator over the results datasets.

        Returns:
            Iterator: An iterator over the results datasets.
        """
        return iter(self.data.items())

    @property
    def experiment(self) -> str:
        """Gets the YAML representation of the executed experiment.

        Returns:
            str: The YAML string of the executed experiment.
        """
        return self._file[ExperimentResults.EXPERIMENT_PATH][()].decode("utf-8")  # type: ignore[index]

    @property
    def platform(self) -> str:
        """Gets the YAML representation of the platform.

        Returns:
            str: The YAML string of the platform.
        """
        return self._file[ExperimentResults.PLATFORM_PATH][()].decode("utf-8")

    @property
    def executed_at(self) -> datetime:
        """Gets the timestamp when execution of the experiment started.

        Returns:
            datetime: The timestamp of when the experiment started.
        """
        return datetime.strptime(
            self._file[ExperimentResults.EXECUTED_AT_PATH][()].decode("utf-8"), "%Y-%m-%d %H:%M:%S"
        )

    @property
    def execution_time(self) -> float:
        """Gets the execution time in seconds.

        Returns:
            float: The execution time in seconds.
        """
        return float(self._file[ExperimentResults.EXECUTION_TIME_PATH][()].decode("utf-8"))

    # pylint: disable=too-many-statements
    def plot_S21(self, qprogram: int | str = 0, measurement: int | str = 0, save_plot: bool = True):
        """Plots the S21 parameter from the experiment results.

        Args:
            qprogram (int | str, optional): The index or name of the quantum program. Defaults to 0.
            measurement (int | str, optional): The index or name of the measurement. Defaults to 0.

        Raises:
            NotImplementedError: If the data has more than 2 dimensions.
        """

        def decibels(s21: np.ndarray):
            """Convert result values from s21 into dB"""
            return 20 * np.log10(np.abs(s21))

        def plot_1d(s21: np.ndarray, dims: list[DimensionInfo]):
            """Plot 1d"""
            x_labels, x_values = dims[0].labels, dims[0].values

            fig, ax1 = plt.subplots()
            ax1.set_title(self.path)
            ax1.set_xlabel(x_labels[0])
            ax1.set_ylabel(r"$|S_{21}|$")
            ax1.plot(x_values[0], s21, ".")

            if len(x_labels) > 1:
                # Create secondary x-axis
                ax2 = ax1.twiny()

                # Set labels
                ax2.set_xlabel(x_labels[1])
                ax2.set_xlim(min(x_values[1]), max(x_values[1]))

                # Set tick locations
                ax2_ticks = np.linspace(min(x_values[1]), max(x_values[1]), num=6)
                ax2.set_xticks(ax2_ticks)

                # Force scientific notation
                ax2.ticklabel_format(axis="x", style="sci", scilimits=(-3, 3))

            if save_plot:
                folder = os.path.dirname(self.path)
                path = os.path.join(folder, ExperimentResults.S21_PLOT_NAME)
                fig.savefig(path)

            plt.show()

        # pylint: disable=too-many-locals
        def plot_2d(s21: np.ndarray, dims):
            """Plot 2d"""
            x_labels, x_values = dims[0].labels, dims[0].values
            y_labels, y_values = dims[1].labels, dims[1].values

            # Create x and y edge arrays by extrapolating the edges
            x_edges = np.linspace(x_values[0].min(), x_values[0].max(), len(x_values[0]) + 1)
            y_edges = np.linspace(y_values[0].min(), y_values[0].max(), len(y_values[0]) + 1)

            fig, ax1 = plt.subplots()
            ax1.set_title(self.path)
            ax1.set_xlabel(x_labels[0])
            ax1.set_ylabel(y_labels[0])

            # Force scientific notation
            ax1.ticklabel_format(axis="both", style="sci", scilimits=(-3, 3))

            mesh = ax1.pcolormesh(x_edges, y_edges, s21.T, cmap="viridis", shading="auto")
            fig.colorbar(mesh, ax=ax1)

            if len(x_labels) > 1:
                # Create secondary x-axis
                ax2 = ax1.twiny()

                # Set labels
                ax2.set_xlabel(x_labels[1])
                ax2.set_xlim(min(x_values[1]), max(x_values[1]))

                # Set tick locations
                ax2_ticks = np.linspace(min(x_values[1]), max(x_values[1]), num=6)
                ax2.set_xticks(ax2_ticks)

                # Force scientific notation
                ax2.ticklabel_format(axis="x", style="sci", scilimits=(-3, 3))
            if len(y_labels) > 1:
                ax3 = ax1.twinx()
                ax3.set_ylabel(y_labels[1])
                ax3.set_ylim(min(y_values[1]), max(y_values[1]))

                # Set tick locations
                ax3_ticks = np.linspace(min(y_values[1]), max(y_values[1]), num=6)
                ax3.set_xticks(ax3_ticks)

                # Force scientific notation
                ax3.ticklabel_format(axis="y", style="sci", scilimits=(-3, 3))

            if save_plot:
                folder = os.path.dirname(self.path)
                path = os.path.join(folder, "S21.png")
                fig.savefig(path)

            plt.tight_layout()
            plt.show()

        data, dims = self.get(qprogram=qprogram, measurement=measurement)

        # Calculate S21
        s21 = data[..., 0] + 1j * data[..., 1]
        s21 = decibels(s21)

        n_dimensions = len(s21.shape)
        if n_dimensions == 1:
            plot_1d(s21, dims)
        elif n_dimensions == 2:
            plot_2d(s21, dims)
        else:
            raise NotImplementedError("3D and higher dimension plots are not supported yet.")



import pytest
from unittest.mock import MagicMock, patch
from qililab.instruments import Instrument, Instruments
from qililab.instruments.qblox import QbloxQCM, QbloxQRM
from qililab.qprogram.qblox_compiler import AcquisitionData
from qililab.result import Result
from qililab.typings import Parameter
from qililab.result.qprogram import MeasurementResult
from qililab.platform import Bus

@pytest.fixture
def mock_instruments():
    instrument1 = MagicMock(spec=QbloxQCM)
    instrument2 = MagicMock(spec=QbloxQRM)
    type(instrument1).alias = property(lambda self: "qcm")
    type(instrument2).alias = property(lambda self: "qrm")
    instrument1.is_awg.return_value = True
    instrument1.is_adc.return_value = False
    instrument2.is_awg.return_value = True
    instrument2.is_adc.return_value = True
    return [instrument1, instrument2]

@pytest.fixture
def bus(mock_instruments):
    settings = {
        "alias": "bus1",
        "instruments": ["qcm", "qrm"],
        "channels": [0, 0]
    }
    return Bus(settings=settings, platform_instruments=Instruments(elements=mock_instruments))

class TestBus:

    def test_bus_iter(self, bus):
        for i, (instrument, channel) in enumerate(bus):
            assert instrument == bus.instruments[i]
            assert channel == bus.channels[i]

    def test_bus_alias(self, bus):
        assert bus.alias == "bus1"

    def test_bus_instruments(self, bus):
        assert len(bus.instruments) == 2

    def test_bus_channels(self, bus):
        assert len(bus.channels) == 2
        assert bus.channels == [0, 0]

    def test_bus_str(self, bus):
        assert isinstance(str(bus), str)

    def test_bus_properties(self, bus):
        assert bus.delay == 0
        assert bus.distortions == []

    def test_bus_equality(self, bus):
        other_bus = MagicMock(spec=Bus)
        other_bus.__str__.return_value = str(bus)
        assert bus == other_bus

    def test_bus_inequality(self, bus):
        other_bus = MagicMock(spec=Bus)
        other_bus.__str__.return_value = "different_bus"
        assert bus != other_bus

    def test_bus_to_dict(self, bus):
        expected_dict = {
            "alias": "bus1",
            "instruments": ["qcm", "qrm"],
            "channels": [0, 0]
        }
        assert bus.to_dict() == expected_dict

    def test_bus_has_awg(self, bus):
        assert bus.has_awg() is True

    def test_bus_has_adc(self, bus):
        assert bus.has_adc() is True

    def test_bus_set_parameter(self, bus):
        parameter = MagicMock()
        value = 5
        bus.set_parameter(parameter, value)
        bus.instruments[0].set_parameter.assert_called_once()

    def test_bus_set_parameter_raises_error(self, bus):
        bus.settings.instruments = []
        bus.settings.channels = []
        with pytest.raises(Exception):
            bus.set_parameter(MagicMock(), 5)

    def test_bus_get_parameter(self, bus):
        parameter = MagicMock()
        bus.get_parameter(parameter)
        bus.instruments[0].get_parameter.assert_called_once()

    def test_bus_get_parameter_raises_error(self, bus):
        bus.settings.instruments = []
        bus.settings.channels = []
        with pytest.raises(Exception):
            bus.get_parameter(MagicMock())

    def test_bus_upload_qpysequence(self, bus):
        qpysequence = MagicMock()
        bus.upload_qpysequence(qpysequence)
        bus.instruments[0].upload_qpysequence.assert_called_once()

    def test_bus_upload_qpysequence_raises_error(self, bus):
        bus.settings.instruments = []
        bus.settings.channels = []
        with pytest.raises(AttributeError):
            bus.upload_qpysequence(MagicMock())

    def test_bus_upload(self, bus):
        bus.upload()
        bus.instruments[0].upload.assert_called_once()

    def test_bus_run(self, bus):
        bus.run()
        bus.instruments[0].run.assert_called_once()

    def test_bus_acquire_result(self, bus):
        result = MagicMock(spec=Result)
        bus.instruments[1].acquire_result.return_value = result
        assert bus.acquire_result() == result

    def test_bus_acquire_result_raises_error(self, bus):
        bus.instruments[1].acquire_result.return_value = None
        with pytest.raises(AttributeError, match=f"The bus {bus.alias} cannot acquire results."):
            bus.acquire_result()

        bus.settings.instruments.append(bus.settings.instruments[1])
        result = MagicMock(spec=Result)
        bus.instruments[1].acquire_result.return_value = result
        bus.instruments[2].acquire_result.return_value = result
        with pytest.raises(ValueError, match="Acquisition from multiple instruments is not supported. Obtained a total of 2 results."):
            bus.acquire_result()

    def test_bus_acquire_qprogram_results(self, bus):
        acquisitions = {"acq1": MagicMock(spec=AcquisitionData)}
        results = [MagicMock(spec=MeasurementResult)]
        bus.instruments[1].acquire_qprogram_results.return_value = results
        assert bus.acquire_qprogram_results(acquisitions) == results

    def test_bus_acquire_qprogram_results_raises_error(self, bus):
        acquisitions = {"acq1": MagicMock(spec=AcquisitionData)}
        bus.instruments.clear()
        bus.channels.clear()

        with pytest.raises(AttributeError, match=f"The bus {bus.alias} cannot acquire results because it doesn't have a readout system control."):
            _ = bus.acquire_qprogram_results(acquisitions)

    def test_bus_with_non_existant_instrument_raises_error(self, mock_instruments):
        with pytest.raises(NameError):
            settings = {
                "alias": "bus1",
                "instruments": ["not_in_instruments"],
                "channels": [None, None]
            }
            _ = Bus(settings=settings, platform_instruments=Instruments(elements=mock_instruments))



import re
from dataclasses import asdict
from unittest.mock import MagicMock, patch

import numpy as np
import pytest
import qibo
from qibo import gates
from qibo.backends import NumpyBackend
from qibo.gates import CZ, M, X
from qibo.models import Circuit
import networkx as nx

from qililab.digital import CircuitTranspiler
from qililab.digital.native_gates import Drag, Wait
from qililab.pulse import PulseSchedule
from qililab.settings.digital import DigitalCompilationSettings
from qililab.digital import DigitalTranspilationConfig

qibo.set_backend("numpy")  # set backend to numpy (this is the faster option for < 15 qubits)

# transpilable gates
default_gates = [
    gates.I(0),
    gates.X(0),
    gates.Y(0),
    gates.Z(0),
    gates.H(0),
    gates.RX(0, 0),
    gates.RY(0, 0),
    gates.RZ(0, 0),
    gates.U1(0, 0),
    gates.U2(0, 0, 0),
    gates.U3(0, 0, 0, 0),
    gates.S(0),
    gates.SDG(0),
    gates.T(0),
    gates.TDG(0),
    gates.CNOT(0, 1),
    gates.CZ(0, 1),
    gates.SWAP(0, 1),
    gates.iSWAP(0, 1),
    gates.CRX(0, 1, 0),
    gates.CRY(0, 1, 0),
    gates.CRZ(0, 1, 0),
    gates.CU1(0, 1, 0),
    gates.CU2(0, 1, 0, 0),
    gates.CU3(0, 1, 0, 0, 0),
    gates.FSWAP(0, 1),
    gates.RXX(0, 1, 0),
    gates.RYY(0, 1, 0),
    gates.RZZ(0, 1, 0),
    gates.TOFFOLI(0, 1, 2),
]


def random_circuit(
    nqubits: int,
    ngates: int,
    rng: np.random.Generator,
    circuit_gates: list[qibo.gates.Gate] | None = None,
    exhaustive=False,
) -> Circuit:
    """Generates random qibo circuit with ngates

    Args:
        nqubits (int): number of qubits in the circuit
        ngates (int): number of gates in the circuit
        circuit_gates (dict[gates:int]): dictionary with gates and amount of qubits where those should be applied
        exhaustive (bool) : use all gates at least once (requires ngates>=len(gates))

    Returns:
        c (qibo.Circuit) : resulting circuit
    """

    # get list available gates
    if circuit_gates is None:
        circuit_gates = default_gates

    # init circuit
    c = Circuit(nqubits)

    # get list of gates to use
    if not exhaustive:
        list_gates = rng.choice(circuit_gates, ngates)
    # if exhaustive = True then add all the gates available
    else:
        if ngates < len(circuit_gates):
            raise ValueError("If exhaustive is set to True then ngates must be bigger than len(circuit_gates)!")
        list_gates = []
        for _ in range(ngates // len(circuit_gates)):
            list_gates.extend(circuit_gates)
        list_gates.extend(rng.choice(circuit_gates, ngates % len(circuit_gates), replace=False))
        rng.shuffle(list_gates)

    # add gates iteratively
    for gate in list_gates:
        # apply gate to random qubits
        new_qubits = rng.choice(list(range(nqubits)), len(gate.qubits), replace=False)
        gate = gate.on_qubits(dict(enumerate(new_qubits)))
        if (len(gate.parameters) != 0) and gate.name != "id":
            new_params = tuple(1 for _ in range(len(gate.parameters)))
            gate.parameters = new_params
        c.add(gate)

    return c


def apply_circuit(circuit: Circuit) -> np.ndarray:
    """Apply native gates from a transpiled circuit
    Drag pulses are translated to supported qibo.gates RX, RZ.
    Gates are applied onto initial state |00...0>
    Backend is set to numpy for faster execution since circuits are small

    Args:
        circuit (Circuit): transpiled qibo circuit to be executed

    Returns:
        state (np.ndarray): resulting state
    """

    backend = NumpyBackend()

    nqubits = circuit.nqubits
    # start initial state |00...0>
    state = np.zeros(2**nqubits)
    state[0] = 1
    for gate in circuit.queue:
        if gate.name == "rz":
            state = backend.apply_gate(gate, state, nqubits)
        if gate.name == "drag":
            theta = gate.parameters[0]
            phi = gate.parameters[1]
            qubit = gate.qubits[0]
            # apply Drag gate as RZ and RX gates
            state = backend.apply_gate(gates.RZ(qubit, -phi), state, nqubits)
            state = backend.apply_gate(gates.RX(qubit, theta), state, nqubits)
            state = backend.apply_gate(gates.RZ(qubit, phi), state, nqubits)

        if gate.name == "cz":
            state = backend.apply_gate(gate, state, nqubits)

    return state


def compare_circuits(circuit_q: Circuit, circuit_t: Circuit, nqubits: int) -> float:
    """Runs same circuit using transpiled gates and qibo gates,
    and calculates the scalar product of the 2 resulting states

    Args:
        circuit_q (Circuit): qibo circuit (not transpiled)
        circuit_t (Circuit): transpiled qibo circuit
        nqubits (int): number of qubits in the circuit

    Returns:
        float: absolute scalar product of the 2 resulting states from circuit execution
    """

    # get final state
    state_q = circuit_q().state()
    state_t = apply_circuit(circuit_t)

    # if state_t = k*state_q, where k is a global phase
    # then |state_q * state_t| = |state_q * state_q| * |k| = 1
    return np.abs(np.dot(np.conjugate(state_t), state_q))  # type: ignore[no-any-return]


def compare_exp_z(circuit_q: Circuit, circuit_t: Circuit, nqubits: int) -> list[np.ndarray]:
    r"""Runs same circuit using transpiled gates and qibo gates, applies Z operator to all qubits
    and then calculates the modulo of each coefficient of the state vector. This last operation
    removes the phase difference in Z so that if the state vectors have the same Z observables
    then the state vector coefficients will be the same.
    That is, for the state vector psi, psi' for the transpiled and qibo circuits
    .. math:: \Psi_k == \Psi'_k \forall k
    Where each psi_k is the coefficient of each basis of the 2^n dimensional state vector
    representation in Z

    Args:
        circuit_q (Circuit): qibo circuit (not transpiled)
        circuit_t (Circuit): transpiled qibo circuit
        nqubits (int): number of qubits in the circuit

    Returns:
        list[np.ndarray]: a list with 2 arrays, one corresponding to the state vector for the
            transpiled circuit and another one for the qibo circuit
    """

    # get final states and save them for modulo calculation
    state_q = circuit_q().state()
    state_q_0 = state_q.copy()
    state_t = apply_circuit(circuit_t)
    state_t_0 = state_t.copy()

    # apply measurement in Z
    backend = NumpyBackend()
    for q in range(circuit_q.nqubits):
        state_q = backend.apply_gate(gates.Z(q), state_q, circuit_q.nqubits)
        state_t = backend.apply_gate(gates.Z(q), state_t, circuit_q.nqubits)

    return [
        np.array([i * k for i, k in zip(np.conjugate(state_t_0), state_t)]),
        np.array([i * k for i, k in zip(np.conjugate(state_q_0), state_q)]),
    ]

@pytest.fixture(name="digital_settings")
def fixture_digital_compilation_settings() -> DigitalCompilationSettings:
    """Fixture that returns an instance of a ``Runcard.GatesSettings`` class."""
    digital_settings_dict = {
        "minimum_clock_time": 5,
        "delay_before_readout": 0,
        "topology": [(0, 2), (1, 2), (2, 3), (2, 4)],
        "gates": {
            "M(0)": [
                {
                    "bus": "readout_q0",
                    "pulse": {
                        "amplitude": 0.8,
                        "phase": 0,
                        "duration": 200,
                        "shape": {"name": "rectangular"},
                    },
                }
            ],
            "Drag(0)": [
                {
                    "bus": "drive_q0",
                    "pulse": {
                        "amplitude": 0.8,
                        "phase": 0,
                        "duration": 198,  # try some non-multiple of clock time (4)
                        "shape": {"name": "drag", "drag_coefficient": 0.8, "num_sigmas": 2},
                    },
                }
            ],
            # random X schedule
            "X(0)": [
                {
                    "bus": "drive_q0",
                    "pulse": {
                        "amplitude": 0.8,
                        "phase": 0,
                        "duration": 200,
                        "shape": {"name": "drag", "drag_coefficient": 0.8, "num_sigmas": 2},
                    },
                },
                {
                    "bus": "flux_q0",
                    "wait_time": 30,
                    "pulse": {
                        "amplitude": 0.8,
                        "phase": 0,
                        "duration": 200,
                        "shape": {"name": "drag", "drag_coefficient": 0.8, "num_sigmas": 2},
                    },
                },
                {
                    "bus": "drive_q0",
                    "pulse": {
                        "amplitude": 0.8,
                        "phase": 0,
                        "duration": 100,
                        "shape": {"name": "rectangular"},
                    },
                },
                {
                    "bus": "drive_q4",
                    "pulse": {
                        "amplitude": 0.8,
                        "phase": 0,
                        "duration": 100,
                        "shape": {"name": "gaussian", "num_sigmas": 4},
                    },
                },
            ],
            "M(1)": [
                {
                    "bus": "readout_q1",
                    "pulse": {
                        "amplitude": 0.8,
                        "phase": 0,
                        "duration": 200,
                        "shape": {"name": "rectangular"},
                    },
                }
            ],
            "M(2)": [
                {
                    "bus": "readout_q2",
                    "pulse": {
                        "amplitude": 0.8,
                        "phase": 0,
                        "duration": 200,
                        "shape": {"name": "rectangular"},
                    },
                }
            ],
            "M(3)": [
                {
                    "bus": "readout_q3",
                    "pulse": {
                        "amplitude": 0.7,
                        "phase": 0.5,
                        "duration": 100,
                        "shape": {"name": "gaussian", "num_sigmas": 2},
                    },
                }
            ],
            "M(4)": [
                {
                    "bus": "readout_q4",
                    "pulse": {
                        "amplitude": 0.7,
                        "phase": 0.5,
                        "duration": 100,
                        "shape": {"name": "gaussian", "num_sigmas": 2},
                    },
                }
            ],
            "CZ(2,3)": [
                {
                    "bus": "flux_q2",
                    "wait_time": 10,
                    "pulse": {
                        "amplitude": 0.7,
                        "phase": 0,
                        "duration": 90,
                        "shape": {"name": "snz", "b": 0.5, "t_phi": 1},
                    },
                },
                # park pulse
                {
                    "bus": "flux_q0",
                    "pulse": {
                        "amplitude": 0.7,
                        "phase": 0,
                        "duration": 100,
                        "shape": {"name": "rectangular"},
                    },
                },
            ],
            # test couplers
            "CZ(4, 0)": [
                {
                    "bus": "flux_c2",
                    "wait_time": 10,
                    "pulse": {
                        "amplitude": 0.7,
                        "phase": 0,
                        "duration": 90,
                        "shape": {"name": "snz", "b": 0.5, "t_phi": 1},
                    },
                },
                {
                    "bus": "flux_q0",
                    "pulse": {
                        "amplitude": 0.7,
                        "phase": 0,
                        "duration": 100,
                        "shape": {"name": "rectangular"},
                    },
                },
            ],
            "CZ(0,1)": [
                {
                    "bus": "flux_q1",
                    "pulse": {
                        "amplitude": 0.8,
                        "phase": 0,
                        "duration": 200,
                        "shape": {"name": "rectangular"},
                        "options": {"q0_phase_correction": 1, "q1_phase_correction": 2},
                    },
                }
            ],
            "CZ(0,2)": [
                {
                    "bus": "flux_q2",
                    "pulse": {
                        "amplitude": 0.8,
                        "phase": 0,
                        "duration": 200,
                        "shape": {"name": "rectangular"},
                        "options": {"q1_phase_correction": 2, "q2_phase_correction": 0},
                    },
                }
            ],
        },
        "buses": {
            "readout_q0": {
                "line": "readout",
                "qubits": [0]
            },
            "readout_q1": {
                "line": "readout",
                "qubits": [1]
            },
            "readout_q2": {
                "line": "readout",
                "qubits": [2]
            },
            "readout_q3": {
                "line": "readout",
                "qubits": [3]
            },
            "readout_q4": {
                "line": "readout",
                "qubits": [4]
            },
            "drive_q0": {
                "line": "drive",
                "qubits": [0]
            },
            "drive_q1": {
                "line": "drive",
                "qubits": [1]
            },
            "drive_q2": {
                "line": "drive",
                "qubits": [2]
            },
            "drive_q3": {
                "line": "drive",
                "qubits": [3]
            },
            "drive_q4": {
                "line": "drive",
                "qubits": [4]
            },
            "flux_q0": {
                "line": "flux",
                "qubits": [0]
            },
            "flux_q1": {
                "line": "flux",
                "qubits": [1]
            },
            "flux_q2": {
                "line": "flux",
                "qubits": [2]
            },
            "flux_q3": {
                "line": "flux",
                "qubits": [3]
            },
            "flux_q4": {
                "line": "flux",
                "qubits": [4]
            },
            "flux_c2": {
                "line": "flux",
                "qubits": [2]
            }
        }
    }
    return DigitalCompilationSettings(**digital_settings_dict)  # type: ignore[arg-type]


def get_bus_schedule(pulse_bus_schedule: dict, port: str) -> list[dict]:
    """Helper function for bus schedule data"""

    return [
        {**asdict(schedule)["pulse"], "start_time": schedule.start_time, "qubit": schedule.qubit}
        for schedule in pulse_bus_schedule[port]
    ]


class TestCircuitTranspiler:
    """Tests for the circuit transpiler class"""

    def test_circuit_to_native(self):
        """Tests the circuit to native gates function without/with optimization
        Does not test phase corrections.
        Test that the transpiled circuit outputs same result if
        circuits are the same, and different results if circuits are
        not the same. State vectors should be the same up to a global
        phase difference.
        With optimization, virtual Zs are applied so there will be a phase difference
        between each single qubit. To check that the transpiled and untranspiled circuits
        are equal, we then check the expected value for each element of the state vector
        """
        # FIXME: do these equality tests for the unitary matrix resulting from the circuit rather
        # than from the state vectors for a more full-proof test
        transpiler = CircuitTranspiler(settings=MagicMock())

        # Test with optimizer=False
        rng = np.random.default_rng(seed=42)  # init random number generator

        # circuits are the same
        for _ in range(500):
            nqubits = np.random.randint(4, 10)
            c1 = random_circuit(
                nqubits=nqubits,
                ngates=len(default_gates),
                rng=rng,
                circuit_gates=None,
                exhaustive=True,
            )

            c2_native_gates = transpiler.gates_to_native(c1.queue)

            # check that both c1, c2 are qibo.Circuit
            assert isinstance(c1, Circuit)
            assert isinstance(c2_native_gates, list)
            assert isinstance(c2_native_gates[0], gates.Gate)

            # Build circuit for comparison:
            c2 = Circuit(nqubits)
            [c2.add(gate) for gate in c2_native_gates]

            # check that states are equivalent up to a global phase
            assert np.allclose(1, compare_circuits(c1, c2, nqubits))

        # test with optimizer=True
        rng = np.random.default_rng(seed=42)  # init random number generator

        # circuits are the same
        for _ in range(500):
            nqubits = np.random.randint(4, 10)
            c1 = random_circuit(
                nqubits=nqubits,
                ngates=len(default_gates),
                rng=rng,
                circuit_gates=None,
                exhaustive=True,
            )
            c2_native_gates = transpiler.gates_to_native(c1.queue)
            # check that both c1, c2 are qibo.Circuit
            assert isinstance(c1, Circuit)
            assert isinstance(c2_native_gates, list)
            assert isinstance(c2_native_gates[0], gates.Gate)

            # Build circuit for comparison:
            c2 = Circuit(nqubits)
            [c2.add(gate) for gate in c2_native_gates]

            # check that states have the same absolute coefficients
            z1_exp, z2_exp = compare_exp_z(c1, c2, nqubits)
            assert np.allclose(z1_exp, z2_exp)

    def test_add_phases_from_RZs_and_CZs_to_drags(self, digital_settings):
        """Test that add_phases_from_RZs_and_CZs_to_drags behaves as expected"""
        transpiler = CircuitTranspiler(settings=digital_settings)

        # gate list to optimize
        test_gates = [
            Drag(0, 1, 1),
            gates.CZ(0, 1),
            gates.RZ(1, 1),
            gates.M(0),
            gates.RZ(0, 2),
            Drag(0, 3, 3),
            gates.CZ(0, 2),
            gates.CZ(1, 0),
            Drag(1, 2, 3),
            gates.RZ(1, 0),
        ]
        # resulting gate list from optimization
        result_gates = [
            Drag(0, 1, 1),
            gates.CZ(0, 1),
            gates.M(0),
            Drag(0, 3, 0),
            gates.CZ(0, 2),
            gates.CZ(1, 0),
            Drag(1, 2, -2),
        ]

        # create circuit to test function with
        circuit = Circuit(3)
        circuit.add(test_gates)

        # check that lists are the same
        optimized_gates = transpiler.add_phases_from_RZs_and_CZs_to_drags(circuit.queue, circuit.nqubits)
        for gate_r, gate_opt in zip(result_gates, optimized_gates):
            assert gate_r.name == gate_opt.name
            assert gate_r.parameters == gate_opt.parameters
            assert gate_r.qubits == gate_opt.qubits

    def test_circuit_to_pulses(self, digital_settings):
        """Test translate method"""
        transpiler = CircuitTranspiler(settings=digital_settings)
        # test circuit
        circuit = Circuit(5)
        circuit.add(X(0))
        circuit.add(Drag(0, 1, 0.5))
        circuit.add(CZ(3, 2))
        circuit.add(M(0))
        circuit.add(CZ(2, 3))
        circuit.add(CZ(4, 0))
        circuit.add(M(*range(4)))
        circuit.add(Wait(0, t=10))
        circuit.add(Drag(0, 2, 0.5))

        pulse_schedule = transpiler.gates_to_pulses(circuit.queue)

        # test general properties of the pulse schedule
        assert isinstance(pulse_schedule, PulseSchedule)

        # there are 6 different buses + 3 empty for unused flux lines
        assert len(pulse_schedule) == 12
        assert all(len(schedule_element.timeline) == 0 for schedule_element in pulse_schedule.elements[-3:])

        # we can ignore empty elements from here on
        pulse_schedule.elements = [element for element in pulse_schedule.elements if element.timeline]

        # extract pulse events per bus and separate measurement pulses
        pulse_bus_schedule = {
            pulse_bus_schedule.bus_alias: pulse_bus_schedule.timeline for pulse_bus_schedule in pulse_schedule
        }

        # TODO: I have tested it manually, should add assertions here.


    def test_normalize_angle(self, digital_settings):
        """Test that the angle is normalized properly for drag pulses"""
        c = Circuit(1)
        c.add(Drag(0, 2 * np.pi + 0.1, 0))
        transpiler = CircuitTranspiler(settings=digital_settings)
        pulse_schedule = transpiler.gates_to_pulses(c.queue)
        assert np.allclose(pulse_schedule.elements[0].timeline[0].pulse.amplitude, 0.1 * 0.8 / np.pi)
        c = Circuit(1)
        c.add(Drag(0, np.pi + 0.1, 0))
        transpiler = CircuitTranspiler(settings=digital_settings)
        pulse_schedule = transpiler.gates_to_pulses(c.queue)
        assert np.allclose(pulse_schedule.elements[0].timeline[0].pulse.amplitude, abs(-0.7745352091052967))

    def test_negative_amplitudes_add_extra_phase(self, digital_settings):
        """Test that transpiling negative amplitudes results in an added PI phase."""
        c = Circuit(1)
        c.add(Drag(0, -np.pi / 2, 0))
        transpiler = CircuitTranspiler(settings=digital_settings)
        pulse_schedule = transpiler.gates_to_pulses(c.queue)
        assert np.allclose(pulse_schedule.elements[0].timeline[0].pulse.amplitude, (np.pi / 2) * 0.8 / np.pi)
        assert np.allclose(pulse_schedule.elements[0].timeline[0].pulse.phase, 0 + np.pi)

    def test_drag_schedule_error(self, digital_settings):
        """Test error is raised if len(drag schedule) > 1"""
        # append schedule of M(0) to Drag(0) so that Drag(0)'s gate schedule has 2 elements
        digital_settings.gates["Drag(0)"].append(digital_settings.gates["M(0)"][0])
        gate_schedule = digital_settings.gates["Drag(0)"]
        error_string = re.escape(
            f"Schedule for the drag gate is expected to have only 1 pulse but instead found {len(gate_schedule)} pulses"
        )
        circuit = Circuit(1)
        circuit.add(Drag(0, 1, 1))
        transpiler = CircuitTranspiler(settings=digital_settings)
        with pytest.raises(ValueError, match=error_string):
            transpiler.gates_to_pulses(circuit.queue)


    @pytest.mark.parametrize("optimize", [True, False])
    @patch("qililab.digital.circuit_transpiler.CircuitTranspiler.add_phases_from_RZs_and_CZs_to_drags")
    @patch("qililab.digital.circuit_transpiler.CircuitTranspiler.optimize_gates")
    @patch("qililab.digital.circuit_transpiler.CircuitTranspiler.optimize_transpiled_gates")
    @patch("qililab.digital.circuit_transpiler.CircuitTranspiler.route_circuit")
    @patch("qililab.digital.circuit_transpiler.CircuitTranspiler.gates_to_native")
    @patch("qililab.digital.circuit_transpiler.CircuitTranspiler.gates_to_pulses")
    def test_transpile_circuit(self, mock_to_pulses, mock_to_native, mock_route, mock_opt_trans, mock_opt_circuit, mock_add_phases, optimize, digital_settings):
        """Test transpile_circuit method"""
        transpiler = CircuitTranspiler(settings=digital_settings)
        routing=True
        placer = MagicMock()
        router = MagicMock()
        routing_iterations = 7
        transpilation_config = DigitalTranspilationConfig(routing=routing, optimize=optimize, router=router, placer=placer, routing_iterations=routing_iterations)

        # Mock circuit for return values
        mock_circuit = Circuit(5)
        mock_circuit.add(Drag(0, 2*np.pi, np.pi))
        mock_circuit_gates = mock_circuit.queue

        # Mock layout for return values
        mock_layout = [0, 1, 2, 3, 4]

        # Mock schedule for return values
        mock_schedule = PulseSchedule()

        # Mock the return values
        mock_route.return_value = mock_circuit.queue, mock_circuit.nqubits, mock_layout
        mock_opt_circuit.return_value = mock_circuit_gates
        mock_to_native.return_value = mock_circuit_gates
        mock_add_phases.return_value = mock_circuit_gates
        mock_opt_trans.return_value = mock_circuit_gates
        mock_to_pulses.return_value = mock_schedule

        circuit = random_circuit(5, 10, np.random.default_rng())

        schedule, layout = transpiler.transpile_circuit(circuit, transpilation_config)

        # Mandatory asserts in order:
        mock_route.assert_called_once_with(circuit, placer, router, routing_iterations)
        mock_to_native.assert_called_once_with(mock_circuit.queue)
        mock_add_phases.assert_called_once_with(mock_circuit_gates, mock_circuit.nqubits)
        mock_to_pulses.assert_called_once_with(mock_circuit_gates)
        assert (schedule, layout) == (mock_schedule, mock_layout)

        # Asserts if optimize=True:
        if optimize:
            mock_opt_circuit.assert_called_once_with(mock_circuit_gates)
            mock_opt_trans.assert_called_once_with(mock_circuit_gates)
        else:
            mock_opt_circuit.assert_not_called()
            mock_opt_trans.assert_not_called()

            # Test if routing skipped:
            transpilation_config.routing = False
            mock_route.reset_mock()
            _, _ = transpiler.transpile_circuit(circuit, transpilation_config)
            mock_route.assert_not_called()

            # Test if no config is provided:
            mock_route.reset_mock()
            _, _ = transpiler.transpile_circuit(circuit)
            mock_route.assert_not_called()
            mock_opt_circuit.assert_not_called()
            mock_opt_trans.assert_not_called()


    @patch("qililab.digital.circuit_router.CircuitRouter.route")
    def test_route_trivial_circuit(self, mock_route, digital_settings):
        """Test route_circuit method"""
        transpiler = CircuitTranspiler(settings=digital_settings)
        routing_iterations = 7

        # Mock the return values
        mock_circuit = Circuit(5)
        mock_circuit.add(X(0))
        mock_layout = [0, 1, 2, 3, 4]
        mock_route.return_value = (mock_circuit, mock_layout)

        # Execute the function
        circuit_gates, nqubits, layout = transpiler.route_circuit(mock_circuit, iterations=routing_iterations)

        # Asserts:
        mock_route.assert_called_once_with(mock_circuit, routing_iterations)
        assert (circuit_gates, nqubits, layout) == (mock_circuit.queue, mock_circuit.nqubits, mock_layout)

    def test_route_circuit_only_needs_remapping_integration(self, digital_settings):
        """Test route_circuit method"""
        transpiler = CircuitTranspiler(settings=digital_settings)
        routing_iterations = 10

        # Mock the return values
        mock_circuit = Circuit(5)
        mock_circuit.add(gates.CNOT(1, 0))
        mock_circuit.add(gates.CNOT(0, 1))
        mock_circuit.add(gates.CNOT(3, 0))


        # Execute the function
        circuit_gates, nqubits, final_layout = transpiler.route_circuit(mock_circuit, iterations=routing_iterations)
        output_gates = [(gate.name, gate.qubits) for gate in circuit_gates]

        # Asserts:
        expected_layout = [2, 1, 0, 3, 4]
        expected_gates = [('cx', (1, 2)), ('cx', (2, 1)), ('cx', (3, 2))]

        assert (output_gates, nqubits, final_layout) == (expected_gates, mock_circuit.nqubits, expected_layout)

    def test_route_circuit_swap_needed_integration(self, digital_settings):
        """Test route_circuit method"""
        transpiler = CircuitTranspiler(settings=digital_settings)
        routing_iterations = 10

        # Mock the return values
        mock_circuit = Circuit(4)
        mock_circuit.add(gates.CNOT(1, 0))
        mock_circuit.add(gates.CNOT(3, 2))


        # Execute the function
        circuit_gates, nqubits, final_layout = transpiler.route_circuit(mock_circuit, iterations=routing_iterations)
        output_gates = [(gate.name, gate.qubits) for gate in circuit_gates]

        # Asserts:
        expected_gates_and_layout = [
            ([('cx', (2, 0)), ('swap', (2, 3)), ('cx', (2, 1))], [0, 3, 1, 2, 4]),
            ([('cx', (2, 0)), ('swap', (3, 2)), ('cx', (2, 1))], [0, 3, 1, 2, 4]),
            ([('cx', (2, 0)), ('swap', (2, 1)), ('cx', (3, 2))], [0, 1, 2, 3, 4]),
            ([('cx', (2, 0)), ('swap', (1, 2)), ('cx', (3, 2))], [0, 1, 2, 3, 4]),
            ([('cx', (1, 2)), ('swap', (2, 0)), ('cx', (3, 2))], [0, 1, 2, 3, 4]),
            ([('cx', (1, 2)), ('swap', (0, 2)), ('cx', (3, 2))], [0, 1, 2, 3, 4]),
            ([('cx', (1, 2)), ('swap', (2, 3)), ('cx', (2, 0))], [2, 1, 3, 0, 4]),
            ([('cx', (1, 2)), ('swap', (3, 2)), ('cx', (2, 0))], [2, 1, 3, 0, 4]),
            ([('cx', (3, 2)), ('swap', (2, 0)), ('cx', (1, 2))], [2, 1, 0, 3, 4]),
            ([('cx', (3, 2)), ('swap', (0, 2)), ('cx', (1, 2))], [2, 1, 0, 3, 4]),
            ([('cx', (3, 2)), ('swap', (2, 1)), ('cx', (2, 0))], [0, 2, 1, 3, 4]),
            ([('cx', (3, 2)), ('swap', (1, 2)), ('cx', (2, 0))], [0, 2, 1, 3, 4]),

        ]

        assert nqubits == 5 # The routing changes the size to fit that of the topology
        # Test one of the possible routing has been achieved:
        assert (output_gates, final_layout) in expected_gates_and_layout

    def test_route_circuit_complex_integration(self, digital_settings):
        """Test route_circuit method"""
        transpiler = CircuitTranspiler(settings=digital_settings)
        routing_iterations = 10

        # Mock the return values
        mock_circuit = Circuit(5)
        mock_circuit.add(gates.CNOT(0, 1))
        mock_circuit.add(gates.CNOT(1, 2))
        mock_circuit.add(gates.CNOT(3, 0))
        mock_circuit.add(gates.CNOT(1, 4))
        mock_circuit.add(gates.CNOT(3, 4))
        mock_circuit.add(gates.CNOT(1, 0))


        # Execute the function
        circuit_gates, _, final_layout = transpiler.route_circuit(mock_circuit, iterations=routing_iterations)
        output_gates = [[gate.name, gate.qubits] for gate in circuit_gates]

        # Undo routing with SWAPS
        for idx, gate in enumerate(output_gates):
            if gate[0] == 'swap':
                output_gates[idx] = None
                mapping_to_do = {gate[1][0]: gate[1][1], gate[1][1]:gate[1][0]}
                for idx_2, gate_2 in enumerate(output_gates[idx+1:]):
                    qubit_1, qubit_2 = gate_2[1]
                    output_gates[idx_2+idx+1][1] = (mapping_to_do.get(qubit_1, qubit_1), mapping_to_do.get(qubit_2, qubit_2))
                # Change layout:
                final_layout[gate[1][0]], final_layout[gate[1][1]] = final_layout[gate[1][1]], final_layout[gate[1][0]]

        # Undo initial mapping
        for idx, gate in enumerate(output_gates):
            if gate is not None:
                output_gates[idx] = gate[0], tuple(final_layout[q] for q in gate[1])

        output_gates = [gate for gate in output_gates if gate is not None]

        # Test you retrieve original circuit:
        assert sorted(output_gates) == sorted([(gate.name, gate.qubits) for gate in mock_circuit.queue])

    @patch("qililab.digital.circuit_transpiler.nx.Graph")
    @patch("qililab.digital.circuit_transpiler.CircuitRouter")
    def test_that_route_circuit_instantiates_Router(self, mock_router, mock_graph, digital_settings):
        """Test route_circuit method"""
        transpiler = CircuitTranspiler(settings=digital_settings)
        routing_iterations = 7

        # Mock the return values
        mock_circuit = Circuit(5)
        mock_circuit.add(X(0))

        graph_mocking = nx.Graph(transpiler.settings.topology)
        mock_graph.return_value = graph_mocking

        # Execute the function
        with pytest.raises(ValueError, match=re.escape("not enough values to unpack (expected 2, got 0)")):
            transpiler.route_circuit(mock_circuit, iterations=routing_iterations)

        # Asserts:
        mock_router.assert_called_once_with(graph_mocking, None, None)

    def test_DigitalTranspilerConfig(self):
        """Test that the dataclass default values are correct, and that properies give the correct order
        """
        tc = DigitalTranspilationConfig()
        assert tc._attributes_ordered == (False, None, None, 10, False)



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Qblox Scope Data"""

from dataclasses import dataclass


@dataclass
class ScopeData:
    """Scope data."""

    @dataclass
    class PathData:
        """Path data."""

        data: list[float]
        avg_cnt: int
        out_of_range: bool

    path0: PathData
    path1: PathData

    def __post_init__(self):
        """Change invalid name and cast to PathData class."""
        if isinstance(self.path0, dict) and isinstance(self.path1, dict):
            if "out-of-range" in self.path0:
                self.path0["out_of_range"] = self.path0.pop("out-of-range")
            self.path0 = self.PathData(**self.path0)
            if "out-of-range" in self.path1:
                self.path1["out_of_range"] = self.path1.pop("out-of-range")
            self.path1 = self.PathData(**self.path1)



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Constants"""

# Environment variables
DATA = "DATA"  # variable containing the path where data is saved
RUNCARDS = "RUNCARDS"  # variable containing the runcard's path

RESULTS_FILENAME = "results.yml"
EXPERIMENT_FILENAME = "experiment.yml"

DEFAULT_PLATFORM_NAME = "galadriel"

DEFAULT_TIMEOUT = 10 * 1000  # 10 seconds

GATE_ALIAS_REGEX = r"(?P<gate>[a-zA-Z]+)\((?P<qubits>\d+(?:,\s*\d+)*)\)"

FLUX_CONTROL_REGEX = r"^(?P<flux>phi[xzy])_(?:q(?P<qubit>\d+)|c(?P<coupler>\d+_\d+))$"  # regex to identify flux control to bus mapping. Example flux control "phix_c0_1" corresponding to phix control of the coupler between qubits 0 and 1

# TODO: Distribute constants over different classes


class RUNCARD:
    """YAML constants."""

    NAME = "name"
    ALIAS = "alias"
    INSTRUMENT = "instrument"
    INSTRUMENTS = "instruments"
    ELEMENTS = "elements"
    READOUT = "readout"
    GATES_SETTINGS = "gates_settings"
    PLATFORM = "platform"
    BUSES = "buses"
    AWG = "awg"
    SIGNAL_GENERATOR = "signal_generator"
    ATTENUATOR = "attenuator"
    SYSTEM_CONTROL = "system_control"
    INSTRUMENT_CONTROLLER = "instrument_controller"
    INSTRUMENT_CONTROLLERS = "instrument_controllers"
    FIRMWARE = "firmware"
    GATES = "gates"
    VOLTAGE_SOURCE = "voltage_source"
    CURRENT_SOURCE = "current_source"
    DISTORTIONS = "distortions"
    DELAY = "delay"
    FLUX_CONTROL_TOPOLOGY = "flux_control_topology"
    CHANNELS = "channels"
    DIGITAL = "digital"
    ANALOG = "analog"


class PLATFORM:
    """Platform attribute names."""

    DELAY_BETWEEN_PULSES = "delay_between_pulses"
    DELAY_BEFORE_READOUT = "delay_before_readout"
    TIMINGS_CALCULATION_METHOD = "timings_calculation_method"
    RESET_METHOD = "reset_method"
    PASSIVE_RESET_DURATION = "passive_reset_duration"
    MINIMUM_CLOCK_TIME = "minimum_clock_time"


class CURRENTSOURCE:
    """CurrentSource attribute names."""

    CURRENT = "current"
    SPAN = "span"
    RAMPING_ENABLED = "ramping_enabled"
    RAMP_RATE = "ramp_rate"


class VOLTAGESOURCE:
    """VoltageSource attribute names."""

    VOLTAGE = "voltage"
    SPAN = "span"
    RAMPING_ENABLED = "ramping_enabled"
    RAMP_RATE = "ramp_rate"


class SIGNALGENERATOR:
    """SignalGenerator attribute names."""

    FREQUENCY = "frequency"
    PULSES = "pulses"


class EXPERIMENT:
    """Experiment constants."""

    HARDWARE_AVERAGE = "hardware_average"
    SOFTWARE_AVERAGE = "software_average"
    REPETITION_DURATION = "repetition_duration"
    SHAPE = "shape"
    RESULTS = "results"
    NUM_SCHEDULES = "num_schedules"
    LOOPS = "loops"
    OPTIONS = "options"
    CIRCUITS = "circuits"
    PULSE_SCHEDULES = "pulse_schedules"
    DESCRIPTION = "description"


class BUS:
    """Bus constants."""

    PORT = "port"
    SYSTEM_CONTROL = "system_control"
    ATTENUATOR = "attenuator"
    SCHEDULES = "schedules"
    NUM_SCHEDULES = "num_schedules"
    SHAPE = "shape"  # shape of the results
    RESULTS = "results"


class NODE:
    """Chip node class and attribute names"""

    NODES = "nodes"
    FREQUENCY = "frequency"
    QUBIT_INDEX = "qubit_index"
    LINE = "line"


class PULSESCHEDULES:
    """Pulse Schedules attribute names."""

    ELEMENTS = "elements"


class PULSEBUSSCHEDULE:
    """PULSE BUS SCHEDULE attribute names."""

    TIMELINE = "timeline"
    PORT = "port"


class PULSE:
    """Pulse attribute names."""

    NAME = "name"
    AMPLITUDE = "amplitude"
    FREQUENCY = "frequency"
    PHASE = "phase"
    DURATION = "duration"
    PORT = "port"
    PULSE_SHAPE = "pulse_shape"


class PULSESHAPE:
    """PulseShape attribute names."""

    NAME = "name"


class PULSEDISTORTION:
    """PulsePredecessor attribute names."""

    NAME = "name"


class PULSEEVENT:
    """PulseEvent attribute names."""

    PULSE = "pulse"
    PULSE_DISTORTIONS = "pulse_distortions"
    START_TIME = "start_time"


class INSTRUMENTCONTROLLER:
    """Instrument controller attribute names."""

    CONNECTION = "connection"
    MODULES = "modules"
    RESET = "reset"


class CONNECTION:
    """Connection attribute names."""

    ADDRESS = "address"


class QBLOXRESULT:
    """Qblox Results attribute names."""

    INTEGRATION_LENGTHS = "integration_lengths"
    QBLOX_RAW_RESULTS = "qblox_raw_results"


class QBLOXMEASUREMENTRESULT:
    """Qblox Results attribute names."""

    RAW_MEASUREMENT_DATA = "raw_measurement_data"


class QMRESULT:
    """Quantum Machines Results attribute names."""

    I = "i"
    Q = "q"
    ADC1 = "adc1"
    ADC2 = "adc2"


class RESULTSDATAFRAME:
    """Results DataFrame Attributes"""

    SOFTWARE_AVG_INDEX = "software_avg_index"
    SEQUENCE_INDEX = "sequence_index"
    LOOP_INDEX = "loop_index_"
    QUBIT_INDEX = "qubit_index"
    RESULTS_INDEX = "results_index"
    BINS_INDEX = "bins_index"
    SCOPE_INDEX = "scope_index"
    ACQUISITION_INDEX = "acquisition_index"
    P0 = "p0"
    P1 = "p1"
    I = "i"
    Q = "q"
    AMPLITUDE = "amplitude"
    PHASE = "phase"


UNITS = {"frequency": "Hz"}

UNIT_PREFIX = {
    1e-24: "y",  # yocto
    1e-21: "z",  # zepto
    1e-18: "a",  # atto
    1e-15: "f",  # femto
    1e-12: "p",  # pico
    1e-9: "n",  # nano
    1e-6: "u",  # micro
    1e-3: "m",  # mili
    1e-2: "c",  # centi
    1e-1: "d",  # deci
    1e3: "k",  # kilo
    1e6: "M",  # mega
    1e9: "G",  # giga
    1e12: "T",  # tera
    1e15: "P",  # peta
    1e18: "E",  # exa
    1e21: "Z",  # zetta
    1e24: "Y",  # yotta
}


class QBLOXCONSTANTS:
    """Qblox Constants"""

    SCOPE_LENGTH = 16380


class AWGTypes:
    """Typings from AWG Types"""

    AWG_SEQUENCERS = "awg_sequencers"
    OUT_OFFSETS = "out_offsets"


class AWGSequencerTypes:
    """Types from AWG Sequencer Types"""

    IDENTIFIER = "identifier"
    INTERMEDIATE_FREQUENCY = "intermediate_frequency"
    OFFSET_I = "offset_i"
    OFFSET_Q = "offset_q"


class AWGIQChannelTypes:
    """Types from AWG IQ Channel Types"""

    IDENTIFIER = "identifier"
    I_CHANNEL = "i_channel"
    Q_CHANNEL = "q_channel"


class AWGChannelMappingTypes:
    """Types from AWG Channel Mapping Types"""

    AWG_SEQUENCER_IDENTIFIER = "awg_sequencer_identifier"
    AWG_SEQUENCER_PATH_IDENTIFIER = "awg_sequencer_path_identifier"


class AWGOutputChannelTypes:
    """Types from AWG Ouput Channel Types"""

    IDENTIFIER = "identifier"



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module contains the QProgram class and all the needed information to build a QProgram.

.. currentmodule:: qililab

QProgram Class
~~~~~~~~~~~~~~~~

.. autosummary::
    :toctree: api

    ~QProgram

Compilers
~~~~~~~~~

.. autosummary::
    :toctree: api

    ~QbloxCompiler
    ~QuantumMachinesCompiler

Other QProgram related Classes
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

.. autosummary::
    :toctree: api

    ~Calibration
    ~Domain

"""

from .calibration import Calibration
from .crosstalk_matrix import CrosstalkMatrix, FluxVector
from .experiment import Experiment
from .qblox_compiler import QbloxCompilationOutput, QbloxCompiler
from .qprogram import QProgram
from .quantum_machines_compiler import QuantumMachinesCompilationOutput, QuantumMachinesCompiler
from .variable import Domain

__all__ = [
    "Calibration",
    "CrosstalkMatrix",
    "Domain",
    "Experiment",
    "FluxVector",
    "QProgram",
    "QbloxCompilationOutput",
    "QbloxCompiler",
    "QuantumMachinesCompilationOutput",
    "QuantumMachinesCompiler",
]



import os
from unittest.mock import patch

import pytest
from tests.qprogram.test_structured_program import TestStructuredProgram

from qililab.qprogram.calibration import Calibration
from qililab.qprogram.crosstalk_matrix import CrosstalkMatrix
from qililab.qprogram.experiment import Experiment
from qililab.qprogram.operations import ExecuteQProgram, SetParameter
from qililab.qprogram.operations.set_crosstalk import SetCrosstalk
from qililab.qprogram.qprogram import QProgram
from qililab.qprogram.variable import Domain
from qililab.typings.enums import Parameter
from qililab.utils.serialization import deserialize, deserialize_from, serialize, serialize_to
from qililab.waveforms import IQPair, Square


class TestExperiment(TestStructuredProgram):
    """Unit tests checking the QProgram attributes and methods"""

    @pytest.fixture
    def instance(self):
        return Experiment(label="experiment")

    def test_set_parameter(self, instance: Experiment):
        """Test set_awg_gain method"""
        instance.set_parameter(alias="flux_bus", parameter=Parameter.VOLTAGE, value=0.5)

        assert len(instance._active_block.elements) == 1
        assert len(instance._body.elements) == 1
        assert isinstance(instance._body.elements[0], SetParameter)
        assert instance._body.elements[0].alias == "flux_bus"
        assert instance._body.elements[0].parameter == Parameter.VOLTAGE
        assert instance._body.elements[0].value == 0.5

    def test_set_parameter_flux(self, instance: Experiment):
        """Test set_parameter method with FLUX parameter and crosstalk"""
        instance.set_crosstalk(CrosstalkMatrix.from_buses({"flux_bus": {"flux_bus": 1.0}}))
        instance.set_parameter(alias="flux_bus", parameter=Parameter.FLUX, value=0.5)

        assert len(instance._active_block.elements) == 2
        assert len(instance._body.elements) == 2
        assert isinstance(instance._body.elements[1], SetParameter)
        assert isinstance(instance._body.elements[0], SetCrosstalk)
        assert instance._body.elements[1].alias == "flux_bus"
        assert instance._body.elements[1].parameter == Parameter.FLUX
        assert instance._body.elements[1].value == 0.5

    def test_execute_qprogram(self, instance: Experiment):
        """Test execute_qprogram method"""
        qp = QProgram()
        instance.execute_qprogram(qprogram=qp)

        assert len(instance._active_block.elements) == 1
        assert len(instance._body.elements) == 1
        assert isinstance(instance._body.elements[0], ExecuteQProgram)
        assert instance._body.elements[0].qprogram is qp

    def test_serialization_deserialization(self, instance: Experiment):
        """Test serialization and deserialization works."""
        file = "test_serialization_deserialization_experiment.yml"
        qp = QProgram()
        gain = qp.variable(label="gain", domain=Domain.Voltage)
        with qp.for_loop(variable=gain, start=0.0, stop=1.0, step=0.1):
            qp.set_gain(bus="drive_bus", gain=gain)
            qp.play(bus="drive_bus", waveform=IQPair(I=Square(1.0, 200), Q=Square(1.0, 200)))

        voltage = instance.variable(label="voltage", domain=Domain.Voltage)
        with instance.for_loop(variable=voltage, start=0.0, stop=1.0, step=0.1):
            instance.set_parameter(alias="flux_bus", parameter=Parameter.VOLTAGE, value=voltage)
            instance.execute_qprogram(qp)

        serialized = serialize(instance)
        deserialized_experiment = deserialize(serialized, Experiment)

        assert isinstance(deserialized_experiment, Experiment)

        serialize_to(instance, file=file)
        deserialized_experiment = deserialize_from(file, Experiment)

        assert isinstance(deserialized_experiment, Experiment)

        os.remove(file)



""" Test Results """
import os

import numpy as np
import pytest

from qililab.result.qprogram.qblox_measurement_result import QbloxMeasurementResult
from qililab.result.qprogram.qprogram_results import QProgramResults
from qililab.result.qprogram.quantum_machines_measurement_result import QuantumMachinesMeasurementResult
from qililab.utils.serialization import deserialize, deserialize_from, serialize, serialize_to


@pytest.fixture(name="qblox_qprogram_results")
def fixture_qblox_qprogram_results():
    """fixture_qblox_result_noscope

    Returns:
        QProgramResults: `QProgramResults` from Qblox instrument fixture.
    """
    results = QProgramResults()
    results.append_result(
        bus="readout", result=QbloxMeasurementResult(bus="readout", raw_measurement_data={"abc": 123})
    )
    return results


@pytest.fixture(name="quantum_machines_qprogram_results")
def fixture_quantum_machines_qprogram_results():
    """fixture_quantum_machines_result_noscope

    Returns:
        QProgramResults: `QProgramResults` from Quantum Machines instrument fixture.
    """
    results = QProgramResults()
    results.append_result(
        bus="readout",
        result=QuantumMachinesMeasurementResult(bus="readout", I=np.linspace(0, 10, 11), Q=np.linspace(90, 100, 11)),
    )
    return results


class TestsQProgramResult:
    """Test `QbloxQProgramResults` functionalities."""

    def test_append_result_method_with_quantum_machines(self, quantum_machines_qprogram_results: QProgramResults):
        """Tests the instantiation of a QbloxQProgramResult object.

        Args:
            qblox_result_scope (QbloxQProgramResult): QbloxQProgramResult instance.
        """
        assert "readout" in quantum_machines_qprogram_results.results
        assert len(quantum_machines_qprogram_results.results["readout"]) == 1

        quantum_machines_qprogram_results.append_result(
            "readout",
            result=QuantumMachinesMeasurementResult(
                bus="readout", I=np.linspace(0, 10, 11), Q=np.linspace(90, 100, 11)
            ),
        )
        assert len(quantum_machines_qprogram_results.results["readout"]) == 2

        quantum_machines_qprogram_results.append_result(
            "another_readout",
            result=QuantumMachinesMeasurementResult(
                bus="readout", I=np.linspace(0, 10, 11), Q=np.linspace(90, 100, 11)
            ),
        )
        assert "another_readout" in quantum_machines_qprogram_results.results
        assert len(quantum_machines_qprogram_results.results["another_readout"]) == 1

    def test_append_result_method_with_qblox(self, qblox_qprogram_results: QProgramResults):
        """Tests the instantiation of a QbloxQProgramResult object.

        Args:
            qblox_result_scope (QbloxQProgramResult): QbloxQProgramResult instance.
        """
        assert "readout" in qblox_qprogram_results.results
        assert len(qblox_qprogram_results.results["readout"]) == 1

        qblox_qprogram_results.append_result(
            "readout", result=QbloxMeasurementResult(bus="readout", raw_measurement_data={"def": 456})
        )
        assert len(qblox_qprogram_results.results["readout"]) == 2

        qblox_qprogram_results.append_result(
            "another_readout", result=QbloxMeasurementResult(bus="another_readout", raw_measurement_data={"def": 456})
        )
        assert "another_readout" in qblox_qprogram_results.results
        assert len(qblox_qprogram_results.results["another_readout"]) == 1

    def test_serialization_method(
        self, qblox_qprogram_results: QProgramResults, quantum_machines_qprogram_results: QProgramResults
    ):
        """Test serialization and deserialization works."""
        serialized = serialize(qblox_qprogram_results)
        deserialized_qblox_qprogram_results = deserialize(serialized, QProgramResults)

        assert isinstance(deserialized_qblox_qprogram_results, QProgramResults)

        serialize_to(qblox_qprogram_results, file="qblox_qprogram_results.yml")
        deserialized_qblox_qprogram_results = deserialize_from("qblox_qprogram_results.yml", QProgramResults)

        assert isinstance(deserialized_qblox_qprogram_results, QProgramResults)

        os.remove("qblox_qprogram_results.yml")

        serialized = serialize(quantum_machines_qprogram_results)
        deserialized_quantum_machines_qprogram_results = deserialize(serialized, QProgramResults)

        assert isinstance(deserialized_quantum_machines_qprogram_results, QProgramResults)

        serialize_to(qblox_qprogram_results, file="quantum_machines_qprogram_results.yml")
        deserialized_quantum_machines_qprogram_results = deserialize_from(
            "quantum_machines_qprogram_results.yml", QProgramResults
        )

        assert isinstance(deserialized_quantum_machines_qprogram_results, QProgramResults)

        os.remove("quantum_machines_qprogram_results.yml")



""" Module containging new helper functions for envelopes."""
import numpy as np

from qililab.pulse.pulse_shape import PulseShape


def return_envelope(pulse_shape: PulseShape, env_params: dict[str, int]) -> list[np.ndarray]:
    """Function that generates the envelopes for the tests."""
    return pulse_shape.envelope(
        duration=env_params["duration"],
        amplitude=env_params["amplitude"],
        resolution=env_params["resolution"],
    )



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Singleton classes."""

from abc import ABCMeta
from typing import ClassVar


class Singleton(type):
    """Singleton metaclass used to limit to 1 the number of instances of a subclass."""

    _instances: ClassVar[dict] = {}

    def __call__(cls, *args, **kwargs):
        """Create a new instance of cls if it doesn't already exist."""
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class SingletonABC(ABCMeta):
    """ABC singleton metaclass used to limit to 1 the number of instances of a subclass."""

    _instances: ClassVar[dict] = {}

    def __call__(cls, *args, **kwargs):
        """Create a new instance of cls if it doesn't already exist."""
        if cls not in cls._instances:
            cls._instances[cls] = super(SingletonABC, cls).__call__(*args, **kwargs)
        return cls._instances[cls]



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass, field

from qililab.pulse.pulse_distortion import PulseDistortion
from qililab.typings import ChannelID


@dataclass
class BusSettings:
    """Dataclass with all the settings the buses of the platform need.
    Args:
        alias (str): Alias of the bus.
        system_control (dict): Dictionary containing the settings of the system control of the bus.
        distortions (list[dict]): List of dictionaries containing the settings of the distortions applied to each
            bus.
        delay (int, optional): Delay applied to all pulses sent in this bus. Defaults to 0.
    """

    alias: str
    instruments: list[str]
    channels: list[ChannelID | None]
    delay: int = 0
    distortions: list[PulseDistortion] = field(default_factory=list)
